# -*- coding: utf-8 -*-

import abc
import itertools

import numpy as np

from astropy import (
    units,
)

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.py import (
    classname,
    issequenceof,
    float_cast,
    int_cast,
    get_smallest_numpy_int_type,
)
from skyllh.core.random import (
    RandomChoice,
)
from skyllh.core.services import (
    DatasetSignalWeightFactorsService,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)


class SignalGenerator(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """This is the abstract base class for all signal generator classes in
    SkyLLH. It defines the interface for a signal generator.
    """
    def __init__(
            self,
            shg_mgr,
            **kwargs,
    ):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The SourceHypoGroupManager instance defining the source hypothesis
            groups.
        """
        super().__init__(
            **kwargs)

        self.shg_mgr = shg_mgr

    @property
    def shg_mgr(self):
        """The SourceHypoGroupManager instance defining the source hypothesis
        groups.
        """
        return self._shg_mgr

    @shg_mgr.setter
    def shg_mgr(self, manager):
        if not isinstance(manager, SourceHypoGroupManager):
            raise TypeError(
                'The shg_mgr property must be an instance of '
                'SourceHypoGroupManager!')
        self._shg_mgr = manager

    def create_src_params_recarray(
            self,
            src_detsigyield_weights_service):
        """Creates the src_params_recarray structured ndarray of length
        N_sources holding the local source parameter names and values needed for
        the calculation of the detector signal yields.

        Parameters
        ----------
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            The instance of SrcDetSigYieldWeightsService providing the product
            of the source weights with the detector signal yield.

        Returns
        -------
        src_params_recarray : instance of numpy structured ndarray
            The structured numpy ndarray of length N_sources, holding the local
            parameter names and values of each source needed to calculate the
            detector signal yield.
        """
        # Get the parameter names needed for the detector signal yield
        # calculation.
        param_names = []
        for detsigyield in src_detsigyield_weights_service.detsigyield_arr.flat:
            param_names.extend(detsigyield.param_names)
        param_names = set(param_names)

        # Create an empty structured ndarray of length N_sources.
        dt = []
        for pname in param_names:
            dt.extend([
                (pname, np.float64),
                (f'{pname}:gpidx', np.int32)
            ])
        src_params_recarray = np.empty((self._shg_mgr.n_sources,), dtype=dt)

        sidx = 0
        for (shg_idx, shg) in enumerate(self._shg_mgr.shg_list):

            shg_n_src = shg.n_sources

            shg_src_slice = slice(sidx, sidx+shg_n_src)

            pvalues = []
            for pname in param_names:
                pvalues.extend([
                    shg.fluxmodel.get_param(pname),
                    0
                ])

            src_params_recarray[shg_src_slice] = tuple(pvalues)

            sidx += shg_n_src

        return src_params_recarray

    def change_shg_mgr(
            self,
            shg_mgr):
        """Changes the source hypothesis group manager. Derived classes can
        reimplement this method but this method of the base class must still be
        called by the derived class.
        """
        self.shg_mgr = shg_mgr

    @abc.abstractmethod
    def generate_signal_events(
            self,
            rss,
            mean,
            poisson=True,
            src_detsigyield_weights_service=None):
        """This abstract method must be implemented by the derived class to
        generate a given number of signal events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : int | float
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated signal events.
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService | None
            The instance of SrcDetSigYieldWeightsService providing the weighting
            of the sources within the detector. This can be ``None`` if this
            signal generator does not need this information.

        Returns
        -------
        n_signal : int
            The number of generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instances with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        pass


class MultiDatasetSignalGenerator(
        SignalGenerator,
):
    """This is a signal generator class handling multiple datasets by using the
    individual signal generator instances for each dataset. This is the most
    general way to support multiple datasets of different formats and signal
    generation.
    """
    def __init__(
            self,
            shg_mgr,
            dataset_list,
            data_list,
            sig_generator_list=None,
            ds_sig_weight_factors_service=None,
            **kwargs):
        """Constructs a new signal generator handling multiple datasets.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            source hypothesis groups, i.e. the list of sources.
        dataset_list : list of instance of Dataset
            The list of instance of Dataset for which signal events should get
            generated.
        data_list : list of instance of DatasetData
            The list of instance of DatasetData holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        sig_generator_list : list of instance of SignalGenerator | None
            The optional list of instance of SignalGenerator holding
            signal generator instances for each individual dataset. This can be
            ``None`` if this signal generator does not require individual signal
            generators for each dataset.
        ds_sig_weight_factors_service : instance of DatasetSignalWeightFactorsService
            The instance of DatasetSignalWeightFactorsService providing the
            dataset signal weight factor service for calculating the dataset
            signal weights.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            **kwargs)

        self.dataset_list = dataset_list
        self.data_list = data_list
        self.sig_generator_list = sig_generator_list
        self.ds_sig_weight_factors_service = ds_sig_weight_factors_service

        self._src_params_recarray = None

    @property
    def dataset_list(self):
        """The list of Dataset instances for which signal events should get
        generated for.
        """
        return self._dataset_list

    @dataset_list.setter
    def dataset_list(self, datasets):
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The dataset_list property must be a sequence of Dataset '
                'instances!')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the actual data of each
        dataset. The order must match the order of the ``dataset_list``
        property.
        """
        return self._data_list

    @data_list.setter
    def data_list(self, datas):
        if not issequenceof(datas, DatasetData):
            raise TypeError(
                'The data_list property must be a sequence of DatasetData '
                'instances!')
        self._data_list = list(datas)

    @property
    def sig_generator_list(self):
        """The list of instance of SignalGenerator holding signal generator
        instances for each individual dataset.
        """
        return self._sig_generator_list

    @sig_generator_list.setter
    def sig_generator_list(self, generators):
        if generators is not None:
            if not issequenceof(generators, (SignalGenerator, type(None))):
                raise TypeError(
                    'The sig_generator_list property must be a sequence of '
                    'SignalGenerator instances!')
            generators = list(generators)
        self._sig_generator_list = generators

    @property
    def ds_sig_weight_factors_service(self):
        """The instance of DatasetSignalWeightFactorsService providing the
        dataset signal weight factor service for calculating the dataset
        signal weights.
        """
        return self._ds_sig_weight_factors_service

    @ds_sig_weight_factors_service.setter
    def ds_sig_weight_factors_service(self, service):
        if not isinstance(service, DatasetSignalWeightFactorsService):
            raise TypeError(
                'The ds_sig_weight_factors_service property must be an '
                'instance of DatasetSignalWeightFactorsService!')
        self._ds_sig_weight_factors_service = service

    @property
    def n_datasets(self):
        """(read-only) The number of datasets.
        """
        return len(self._dataset_list)

    def change_shg_mgr(
            self,
            shg_mgr):
        """Changes the source hypothesis group manager. This will recreate the
        src_params_recarray needed for calculating the detector signal yields.
        Also it calls the ``change_shg_mgr`` methods of the signal generators of
        the individual datasets.
        """
        super().change_shg_mgr(
            shg_mgr=shg_mgr)

        src_detsigyield_weights_service =\
            self.ds_sig_weight_factors_service.src_detsigyield_weights_service
        self._src_params_recarray = self.create_src_params_recarray(
            src_detsigyield_weights_service=src_detsigyield_weights_service)

        if self.sig_generator_list is None:
            return

        for sig_generator in filter(None, self.sig_generator_list):
            sig_generator.change_shg_mgr(
                shg_mgr=shg_mgr)

    def generate_signal_events(
            self,
            rss,
            mean,
            poisson=True,
            **kwargs):
        """Generates a given number of signal events distributed across the
        individual datasets.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : float | int
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` must be an integer and
            specifies the actual number of generated signal events.

        Returns
        -------
        n_signal : int
            The number of actual generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instances with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        if poisson:
            mean = rss.random.poisson(
                float_cast(
                    mean,
                    'The mean argument must be cast-able to type of float!'))

        mean = int_cast(
            mean,
            'The mean argument must be cast-able to type of int!')

        src_detsigyield_weights_service =\
            self.ds_sig_weight_factors_service.src_detsigyield_weights_service

        # Calculate the dataset weights to distribute the signal events over the
        # datasets.
        if self._src_params_recarray is None:
            self._src_params_recarray = self.create_src_params_recarray(
                src_detsigyield_weights_service=src_detsigyield_weights_service)

        src_detsigyield_weights_service.calculate(
            src_params_recarray=self._src_params_recarray)

        self._ds_sig_weight_factors_service.calculate()
        (ds_weights, _) = self._ds_sig_weight_factors_service.get_weights()

        # Calculate the number of events that need to be generated for each
        # individual dataset. Due to rounding errors, it could happen that the
        # sum of the events is less or greater than ``mean``.
        n_events_arr = np.round(mean * ds_weights, 0).astype(np.int_)

        # Correct the n_events_arr array based on the dataset weights if
        # necessary.
        sum_n_events_arr = np.sum(n_events_arr)
        if sum_n_events_arr != mean:
            (ds_idxs, counts) = np.unique(
                rss.random.choice(
                    np.arange(len(n_events_arr)),
                    size=np.abs(mean - sum_n_events_arr),
                    p=ds_weights,
                ),
                return_counts=True
            )
            if sum_n_events_arr < mean:
                n_events_arr[ds_idxs] += counts
            elif sum_n_events_arr > mean:
                n_events_arr[ds_idxs] -= counts

        n_signal = 0
        signal_events_dict = {}

        for (n_events, ds_sig_generator) in zip(
                n_events_arr,
                self._sig_generator_list):

            (ds_n_signal, ds_sig_events_dict) =\
                ds_sig_generator.generate_signal_events(
                    rss=rss,
                    mean=n_events,
                    poisson=False,
                    src_detsigyield_weights_service=src_detsigyield_weights_service,
                )

            n_signal += ds_n_signal

            for (k, v) in ds_sig_events_dict.items():
                if k not in signal_events_dict:
                    signal_events_dict[k] = v
                else:
                    signal_events_dict[k].append(v)

        return (n_signal, signal_events_dict)


class MCMultiDatasetSignalGenerator(
        MultiDatasetSignalGenerator,
):
    """This is a signal generator class, which handles multiple datasets with
    monte-carlo (MC). It uses the MC events of all datasets to determine the
    possible signal events for a source.
    It does not depend on the detector or source hypothesis, because these
    dependencies are factored out into the signal generation method.
    In fact the construction within this class depends on the construction of
    the signal generation method.
    """
    def __init__(
            self,
            shg_mgr,
            dataset_list,
            data_list,
            valid_event_field_ranges_dict_list=None,
            **kwargs,
    ):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The SourceHypoGroupManager instance defining the source hypothesis
            groups.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        valid_event_field_ranges_dict_list : list of dict | None
            If not ``None``, it specifies for each dataset event fields (key)
            and their valid value range as a 2-element tuple (value). If a
            generated signal event does not fall into a given field range, the
            signal event will be discarded and a new signal event will be drawn.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            dataset_list=dataset_list,
            data_list=data_list,
            **kwargs)

        if valid_event_field_ranges_dict_list is None:
            valid_event_field_ranges_dict_list = [dict()]*len(self.dataset_list)
        if not isinstance(valid_event_field_ranges_dict_list, list):
            raise TypeError(
                'The `valid_event_field_ranges_dict_list` argument must be a list.'
            )
        if len(valid_event_field_ranges_dict_list) != len(self.dataset_list):
            raise ValueError(
                'The valid_event_field_ranges_dict_list argument must be a '
                f'list of length {len(self.dataset_list)}, but it is of length '
                f'{len(valid_event_field_ranges_dict_list)}!')
        self.valid_event_field_ranges_dict_list =\
            valid_event_field_ranges_dict_list

        self._construct_signal_candidates()

    @property
    def valid_event_field_ranges_dict_list(self):
        """The list of dictionary holding the event data fields (key) and their
        valid value range as 2-element tuple (value).
        """
        return self._valid_event_field_ranges_dict_list

    @valid_event_field_ranges_dict_list.setter
    def valid_event_field_ranges_dict_list(self, dict_list):
        if not isinstance(dict_list, list):
            raise TypeError(
                'The valid_event_field_ranges_dict_list must be an instance of '
                'list! '
                f'Its current type is {classname(dict_list)}!')
        for d in dict_list:
            for (k, v) in d.items():
                if not isinstance(k, str):
                    raise TypeError(
                        'Each key of the dictionary of the '
                        'valid_event_field_ranges_dict property must be an '
                        'instance of str! '
                        f'But the type of one of the keys is {classname(k)}!')
                if not isinstance(v, tuple):
                    raise TypeError(
                        'Each value of the dictionary of the '
                        'valid_event_field_ranges_dict property must be an '
                        'instance of tuple! '
                        f'But the value type for the event field "{k}" is '
                        f'{classname(v)}!')
                if len(v) != 2:
                    raise ValueError(
                        f'The tuple for the event field {k} must be of length '
                        f'2! Its current length is {len(v)}!')
        self._valid_event_field_ranges_dict_list = dict_list

    def _construct_signal_candidates(self):
        """Constructs an array holding pointer information of signal candidate
        events pointing into the real MC dataset(s).
        """
        n_datasets = len(self._dataset_list)
        n_sources = self._shg_mgr.n_sources
        shg_list = self._shg_mgr.shg_list
        sig_candidates_dtype = [
            ('ds_idx', get_smallest_numpy_int_type((0, n_datasets))),
            ('ev_idx', get_smallest_numpy_int_type(
                [0]+[len(data.mc) for data in self._data_list])),
            ('shg_idx', get_smallest_numpy_int_type((0, n_sources))),
            ('shg_src_idx', get_smallest_numpy_int_type(
                [0]+[shg.n_sources for shg in shg_list])),
            ('weight', np.float64)
        ]
        self._sig_candidates = np.empty(
            (0,), dtype=sig_candidates_dtype, order='F')

        to_internal_time_unit_factor = self._cfg.to_internal_time_unit(
            time_unit=units.day
        )

        # Go through the source hypothesis groups to get the signal event
        # candidates.
        for ((shg_idx, shg), (j, data)) in itertools.product(
                enumerate(shg_list),
                enumerate(self._data_list)):
            sig_gen_method = shg.sig_gen_method
            if sig_gen_method is None:
                raise ValueError(
                    'No signal generation method has been specified for the '
                    f'source hypothesis group with index {shg_idx}!')
            data_mc = data.mc

            (ev_idx_arr, src_idx_arr, flux_arr) =\
                sig_gen_method.calc_source_signal_mc_event_flux(
                    data_mc=data_mc,
                    shg=shg)

            livetime_days = Livetime.get_integrated_livetime(data.livetime)

            weight = (
                data_mc[ev_idx_arr]['mcweight'] *
                flux_arr *
                livetime_days*to_internal_time_unit_factor
            )

            sig_candidates = np.empty(
                (len(ev_idx_arr),),
                dtype=sig_candidates_dtype,
                order='F'
            )
            sig_candidates['ds_idx'] = j
            sig_candidates['ev_idx'] = ev_idx_arr
            sig_candidates['shg_idx'] = shg_idx
            sig_candidates['shg_src_idx'] = src_idx_arr
            sig_candidates['weight'] = weight

            self._sig_candidates = np.append(
                self._sig_candidates, sig_candidates)
            del sig_candidates

        # Normalize the signal candidate weights.
        self._sig_candidates_weight_sum = np.sum(self._sig_candidates['weight'])
        self._sig_candidates['weight'] /= self._sig_candidates_weight_sum

        # Create new RandomChoice instance for the signal candidates.
        self._sig_candidates_random_choice = RandomChoice(
            items=self._sig_candidates,
            probabilities=self._sig_candidates['weight'])

    def _get_invalid_events_mask(
            self,
            events,
            valid_event_field_ranges_dict,
    ):
        """Determines a boolean mask to select invalid events, which do not
        fulfill the given valid event field ranges.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray of length N_events holding the
            events to check.
        valid_event_field_ranges_dict : dict
            The dictionary holding the data field names (key) and their valid
            value ranges (value).

        Raises
        ------
        KeyError
            If one of the event field does not exist in ``events``.

        Returns
        -------
        mask : instance of numpy.ndarray
            The (N_events,)-shaped numpy.ndarray of bool, holding the mask of
            the invalid events.
        """
        mask = np.zeros((len(events),), dtype=np.bool_)

        for (field_name, min_max) in valid_event_field_ranges_dict.items():
            if field_name not in events:
                raise KeyError(
                    f'The event data field "{field_name}" specified in the '
                    'valid_event_field_ranges_dict does not exist in the event '
                    'data!')
            field_values = events[field_name]
            mask |= (field_values < min_max[0]) | (field_values > min_max[1])

        return mask

    def _draw_valid_sig_events_for_dataset_and_shg(
            self,
            rss,
            mc,
            n_signal,
            ds_idx,
            valid_event_field_ranges_dict,
            shg,
            shg_idx,
    ):
        """Draws n_signal valid signal events for the given dataset and source
        hypothesis group.

        Signal events will be drawn until all events match the event field
        ranges, i.e. are valid signal events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService which should be used to draw
            random numbers from.
        mc : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the monte-carlo events.
        n_signal : int
            The number of signal events to draw.
        ds_idx : int
            The index of the dataset.
        valid_event_field_ranges_dict : dict
            The dictionary holding the data field names (key) and their valid
            value ranges (value) for the requested dataset.
        shg : instance of SourceHypothesisGroup
            The instance of SourceHypothesisGroup for which signal events should
            get drawn.
        shg_idx : int
            The index of the source hypothesis group.

        Returns
        -------
        sig_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the drawn valid signal
            events.
        """
        sig_events = None

        n = 0
        while n < n_signal:
            events_meta = self._sig_candidates_random_choice(
                rss=rss,
                size=n_signal-n,
            )
            m = (events_meta['ds_idx'] == ds_idx) &\
                (events_meta['shg_idx'] == shg_idx)
            events = mc[events_meta['ev_idx'][m]]
            if len(events) > 0:
                events = shg.sig_gen_method.\
                    signal_event_post_sampling_processing(
                        shg, events_meta, events)

                valid_events_mask = np.invert(
                    self._get_invalid_events_mask(
                        events,
                        valid_event_field_ranges_dict,
                    )
                )
                events = events[valid_events_mask]
                if len(events) > 0:
                    if sig_events is None:
                        sig_events = events
                    else:
                        sig_events.append(events)
                    n = len(sig_events)

        return sig_events

    def change_shg_mgr(
            self,
            shg_mgr):
        """Recreates the signal candidates with the changed source hypothesis
        group manager.
        """
        super().change_shg_mgr(
            shg_mgr=shg_mgr)

        self._construct_signal_candidates()

    def mu2flux(
            self,
            mu,
            per_source=False):
        """Translate the mean number of signal events `mu` into the
        corresponding flux. The unit of the returned flux is the internally used
        flux unit.

        Parameters
        ----------
        mu : float
            The mean number of expected signal events for which to get the flux.
        per_source : bool
            Flag if the flux should be returned for each source individually
            (True), or as the sum of all these fluxes (False). The default is
            False.

        Returns
        -------
        mu_flux : float | (n_sources,)-shaped numpy ndarray
            The total flux for all sources (if `per_source = False`) that would
            correspond to the given mean number of detected signal events `mu`.
            If `per_source` is set to True, a numpy ndarray is returned that
            contains the flux for each individual source.
        """
        # Calculate the expected mean number of signal events for each source
        # of the source hypo group manager. For each source we can calculate the
        # flux that would correspond to the given mean number of signal events
        # `mu`. The total flux for all sources is then just the sum.

        # The ref_N variable describes how many total signal events are expected
        # on average for the reference fluxes.
        ref_N = self._sig_candidates_weight_sum

        # The mu_fluxes array is the flux of each source for mu mean detected
        # signal events.
        n_sources = self._shg_mgr.n_sources
        mu_fluxes = np.empty((n_sources,), dtype=np.float64)

        shg_list = self._shg_mgr.shg_list
        mu_fluxes_idx_offset = 0
        for (shg_idx, shg) in enumerate(shg_list):
            fluxmodel = shg.fluxmodel
            # Calculate conversion factor from the flux model unit into the
            # internal flux unit.
            to_internal_flux_unit =\
                fluxmodel.to_internal_flux_unit()
            for k in range(shg.n_sources):
                mask = ((self._sig_candidates['shg_idx'] == shg_idx) &
                        (self._sig_candidates['shg_src_idx'] == k))
                ref_N_k = np.sum(self._sig_candidates[mask]['weight']) * ref_N
                mu_flux_k = (
                    (mu / ref_N) *
                    (ref_N_k / ref_N) *
                    fluxmodel.Phi0 * to_internal_flux_unit)
                mu_fluxes[mu_fluxes_idx_offset + k] = mu_flux_k
            mu_fluxes_idx_offset += shg.n_sources

        if per_source:
            return mu_fluxes

        mu_flux = np.sum(mu_fluxes)
        return mu_flux

    def generate_signal_events(
            self,
            rss,
            mean,
            poisson=True,
            **kwargs):
        """Generates a given number of signal events from the signal candidate
        monte-carlo events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : float | int
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` must be an integer and
            specifies the actual number of generated signal events.

        Returns
        -------
        n_signal : int
            The number of actual generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instances with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        if poisson:
            mean = rss.random.poisson(
                float_cast(
                    mean,
                    'The mean argument must be cast-able to type of float!'))

        n_signal = int_cast(
            mean,
            'The mean argument must be cast-able to type of int!')

        # Draw n_signal signal candidates according to their weight.
        sig_events_meta = self._sig_candidates_random_choice(
            rss=rss,
            size=n_signal,
        )

        # Get the list of unique dataset and source hypothesis group indices of
        # the drawn signal events.
        # Note: This code does not assume the same format for each of the
        #       individual MC datasets, thus might be a bit slower.
        #       If one could assume the same MC dataset format, one
        #       could gather all the MC events of all the datasets first and do
        #       the signal event post processing for all datasets at once.
        signal_events_dict = dict()
        ds_idxs = np.unique(sig_events_meta['ds_idx'])
        for ds_idx in ds_idxs:
            valid_event_field_ranges_dict =\
                self.valid_event_field_ranges_dict_list[ds_idx]
            mc = self._data_list[ds_idx].mc
            ds_mask = sig_events_meta['ds_idx'] == ds_idx
            n_sig_events_ds = np.count_nonzero(ds_mask)

            data = dict([
                (
                    fname,
                    np.empty(
                        (n_sig_events_ds,),
                        dtype=mc.get_field_dtype(fname))
                )
                for fname in mc.field_name_list
            ])
            sig_events = DataFieldRecordArray(data, copy=False)

            fill_start_idx = 0
            # Get the list of unique source hypothesis group indices for the
            # current dataset.
            shg_idxs = np.unique(sig_events_meta[ds_mask]['shg_idx'])
            for shg_idx in shg_idxs:
                shg = self._shg_mgr.shg_list[shg_idx]
                shg_mask = sig_events_meta['shg_idx'] == shg_idx
                # Get the MC events for the drawn signal events.
                ds_shg_mask = ds_mask & shg_mask
                shg_sig_events_meta = sig_events_meta[ds_shg_mask]
                n_shg_sig_events = len(shg_sig_events_meta)
                ev_idx = shg_sig_events_meta['ev_idx']
                # Get the signal MC events of the current dataset and source
                # hypothesis group.
                shg_sig_events = mc[ev_idx]

                # Do the signal event post sampling processing.
                shg_sig_events = shg.sig_gen_method.\
                    signal_event_post_sampling_processing(
                        shg, shg_sig_events_meta, shg_sig_events)

                # Determine the signal events, which do not fulfill the valid
                # event field ranges for this dataset.
                invalid_events_mask = self._get_invalid_events_mask(
                    shg_sig_events,
                    valid_event_field_ranges_dict)
                n_redraw_events = np.count_nonzero(invalid_events_mask)
                if n_redraw_events > 0:
                    # Re-draw n_redraw_events signal events for this dataset
                    # and SHG.
                    redrawn_shg_sig_events =\
                        self._draw_valid_sig_events_for_dataset_and_shg(
                            rss=rss,
                            mc=mc,
                            n_signal=n_redraw_events,
                            ds_idx=ds_idx,
                            valid_event_field_ranges_dict=valid_event_field_ranges_dict,
                            shg=shg,
                            shg_idx=shg_idx,
                        )
                    shg_sig_events[invalid_events_mask] = redrawn_shg_sig_events

                indices = np.indices((n_shg_sig_events,))[0] + fill_start_idx
                sig_events.set_selection(indices, shg_sig_events)

                fill_start_idx += n_shg_sig_events

            signal_events_dict[ds_idx] = sig_events

        return (n_signal, signal_events_dict)



# -*- coding: utf-8 -*-

import logging
import numpy as np

from skyllh.core.utils.analysis import (
    estimate_discovery_potential,
    estimate_sensitivity,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.source_model import (
    PointLikeSource,
)


def generate_ps_sin_dec_h0_ts_values(
        ana, rss, sin_dec_min, sin_dec_max, sin_dec_step, n_bkg_trials=10000,
        n_iter=1, bkg_kwargs=None, ppbar=None):
    """Generates sets of null-hypothesis, i.e. background-only trial data
    events, test-statistic values for the given point-source analysis for a
    grid of sin(dec) values.
    These ts values can the be used for the
    ``estimate_ps_sin_dec_sensitivity_curve`` and
    ``estimate_ps_sin_dec_discovery_potential_curve`` functions.

    Parameters
    ----------
    ana : Analysis instance
        The instance of Analysis to use for generating trials.
    rss : RandomStateService instance
        The instance of RandomStateService to use for generating random numbers
        from.
    sin_dec_min : float
        The minimum sin(dec) value for creating the sin(dec) value grid.
    sin_dec_max : float
        The maximum sin(dec) value for creating the sin(dec) value grid.
    sin_dec_step : float
        The step size in sin(dec) for creating the sin(dec) value grid.
    n_bkg_trials : int, optional
        The number of background trials to generate.
        Default is 10000.
    n_iter : int, optional
        Each set of ts values can be calculated several times to be able to
        estimate the variance of the sensitivity / discovery potential.
        This parameter specifies the number of iterations to perform.
        For each iteration the RandomStateService is re-seeded with a seed that
        is incremented by 1.
        Default is 1.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    ppbar : ProgressBar instance | None
        The optional parent progress bar.

    Returns
    -------
    sin_dec_arr : (n_sin_dec,)-shaped 1D ndarray of float
        The numpy ndarray holding the sin(dec) values for which the ts values
        have been calculated.
    h0_ts_vals_arr : (n_sin_dec,n_iter,n_bkg_trials)-shaped 3D ndarray of float
        The numpy ndarray holding the null-hypothesis ts values for all
        sin(dec) values and iterations.
    """
    logger = logging.getLogger(__name__)

    sin_dec_arr = np.linspace(
        sin_dec_min, sin_dec_max,
        int((sin_dec_max-sin_dec_min)/sin_dec_step)+1, endpoint=True)

    h0_ts_vals_arr = np.empty(
        (len(sin_dec_arr), n_iter, n_bkg_trials), dtype=np.float64)

    logger.debug(
        'Generating %d null-hypothesis trials for %d sin(dec) values, '
        '%d times each, that is %d trials in total',
        n_bkg_trials, len(sin_dec_arr), n_iter,
        n_bkg_trials*len(sin_dec_arr)*n_iter)

    pbar_iter = ProgressBar(n_iter, parent=ppbar).start()
    for iter_idx in range(n_iter):
        pbar_sin_dec = ProgressBar(len(sin_dec_arr), parent=pbar_iter).start()
        for (sin_dec_idx, sin_dec) in enumerate(sin_dec_arr):
            source = PointLikeSource(np.pi, np.arcsin(sin_dec))
            ana.change_source(source)

            h0_ts_vals_arr[sin_dec_idx, iter_idx] = ana.do_trials(
                rss, n_bkg_trials, mean_n_sig=0, bkg_kwargs=bkg_kwargs,
                ppbar=pbar_sin_dec)['ts']

            pbar_sin_dec.increment()
        pbar_sin_dec.finish()

        pbar_iter.increment()

        rss.reseed(rss.seed+1)
    pbar_iter.finish()

    return (sin_dec_arr, h0_ts_vals_arr)


def estimate_ps_sin_dec_sensitivity_curve(
        ana, rss, fitparam_values, sin_dec_arr, h0_ts_vals_arr,
        eps_p=0.0075, mu_min=0, mu_max=20, n_iter=1,
        bkg_kwargs=None, sig_kwargs=None, ppbar=None, **kwargs):
    """Estimates the point-source sensitivity of the given analysis as a
    function of sin(dec). This function places a PointLikeSource source on the
    given declination values and estimates its sensitivity.

    Parameters
    ----------
    ana : Analysis instance
        The instance of Analysis to use for generating trials.
    rss : RandomStateService instance
        The instance of RandomStateService to use for generating random numbers
        from.
    fitparam_values : array_like
        The fit parameter values for which to convert the mean number of signal
        events into a flux.
    sin_dec_arr : (n_sin_dec,)-shaped 1D ndarray
        The ndarray holding the sin(dec) values for which to estimate the
        point-source sensitivity.
    h0_ts_vals_arr : (n_sin_dec,n_iter,n_bkg_trials)-shaped 3D ndarray of float
        The numpy ndarray holding the null-hypothesis ts values for all
        sin(dec) values and iterations.
    eps_p : float, optional
        The precision in probability used for the `estimate_sensitivity`
        function.
        Default is 0.0075.
    mu_min : float, optional
        The minimum value for the mean number of injected signal events as a
        seed for the mu range in which the sensitivity is located.
        Default is 0.
    mu_max : float, optional
        The maximum value for the mean number of injected signal events as a
        seed for the mu range in which the sensitivity is located.
        Default is 20.
    n_iter : int, optional
        Each sensitivity can be estimated several times to be able to estimate
        the variance of the sensitivity. This parameter specifies the number of
        iterations to perform for each sensitivity.
        For each iteration the RandomStateService is re-seeded with a seed that
        is incremented by 1.
        Default is 1.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the given mean number of signal events.
    ppbar : ProgressBar instance | None
        The optional parent progress bar.

    Additional Keyword Arguments
    ----------------------------
    Any additional keyword arguments are passed to the ``estimate_sensitivity``
    function.

    Returns
    -------
    sin_dec_arr : (n_sin_dec,)-shaped 1D ndarray
        The ndarray holding the sin(dec) values for which the sensitivities have
        been estimated.
    mean_ns_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the mean number of signal events corresponding to
        the sensitivity for each sin(dec) value and iteration.
    mean_ns_err_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the estimated error in `mean_ns` for each sin(dec)
        value and iteration.
    flux_scaling_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the scaling factor the reference flux needs to get
        scaled to obtain the flux for the estimated sensitivity.
    """
    logger = logging.getLogger(__name__)

    mu_min_arr = np.repeat(mu_min, len(sin_dec_arr))
    mu_max_arr = np.repeat(mu_max, len(sin_dec_arr))

    mean_ns_arr = np.empty((len(sin_dec_arr), n_iter))
    mean_ns_err_arr = np.empty((len(sin_dec_arr), n_iter))
    flux_scaling_arr = np.empty((len(sin_dec_arr), n_iter))

    pbar_sin_dec = ProgressBar(n_iter, parent=ppbar).start()
    for (sin_dec_idx, sin_dec) in enumerate(sin_dec_arr):
        logger.debug(
            'Estimate point-source sensitivity for sin(dec) = %g, %d times',
            sin_dec, n_iter)
        source = PointLikeSource(np.pi, np.arcsin(sin_dec))
        ana.change_source(source)

        pbar_iter = ProgressBar(len(sin_dec_arr), parent=pbar_sin_dec).start()
        for iter_idx in range(n_iter):
            h0_ts_vals = h0_ts_vals_arr[sin_dec_idx, iter_idx]

            mu_min = mu_min_arr[sin_dec_idx]
            mu_max = mu_max_arr[sin_dec_idx]

            (mean_ns, mean_ns_err) = estimate_sensitivity(
                ana, rss, mu_range=(mu_min, mu_max), eps_p=eps_p,
                h0_ts_vals=h0_ts_vals, bkg_kwargs=bkg_kwargs,
                sig_kwargs=sig_kwargs, ppbar=pbar_iter)

            mean_ns_arr[sin_dec_idx, iter_idx] = mean_ns
            mean_ns_err_arr[sin_dec_idx, iter_idx] = mean_ns_err
            flux_scaling_arr[sin_dec_idx, iter_idx] = ana.calculate_fluxmodel_scaling_factor(
                mean_ns=mean_ns, fitparam_values=np.array(fitparam_values))

            # A new iteration is done, update the mu range using the previous
            # results.
            mu_min_arr = np.mean(mean_ns_arr[:, 0:iter_idx+1]*0.8, axis=1)
            mu_max_arr = np.mean(mean_ns_arr[:, 0:iter_idx+1]*1.2, axis=1)

            rss.reseed(rss.seed+1)

            pbar_iter.increment()
        pbar_iter.finish()

        # It could happen that the first estimation wasn't very accurate due to
        # the unknown seed range for mu. We check for that by calculating the
        # variance of the further iterations and check if the first estimation
        # deviates more than 2 times that variance. If so, recalculate the first
        # estimation.
        if n_iter >= 5:
            mean_ns = mean_ns_arr[sin_dec_idx, 0]
            mean_ns_mean = np.mean(mean_ns_arr[sin_dec_idx, 1:])
            mean_ns_std = np.std(mean_ns_arr[sin_dec_idx, 1:])
            if np.abs(mean_ns - mean_ns_mean) >= 2*mean_ns_std:
                logger.debug(
                    'Detected unprecise estimate for first iteration (mu=%g) '
                    'for sin(dec)=%g: (|%g - %g| >= 2*%g). Recalculating ...',
                    mean_ns, sin_dec, mean_ns, mean_ns_mean, mean_ns_std)
                iter_idx = 0

                h0_ts_vals = h0_ts_vals_arr[sin_dec_idx, iter_idx]

                mu_min = mu_min_arr[sin_dec_idx]
                mu_max = mu_max_arr[sin_dec_idx]

                (mean_ns, mean_ns_err) = estimate_sensitivity(
                    ana, rss, mu_range=(mu_min, mu_max), eps_p=eps_p,
                    h0_ts_vals=h0_ts_vals, bkg_kwargs=bkg_kwargs,
                    sig_kwargs=sig_kwargs, ppbar=pbar_sin_dec)

                mean_ns_arr[sin_dec_idx, iter_idx] = mean_ns
                mean_ns_err_arr[sin_dec_idx, iter_idx] = mean_ns_err
                flux_scaling_arr[sin_dec_idx, iter_idx] = ana.calculate_fluxmodel_scaling_factor(
                    mean_ns=mean_ns, fitparam_values=np.array(fitparam_values))

        pbar_sin_dec.increment()
    pbar_sin_dec.finish()

    return (sin_dec_arr, mean_ns_arr, mean_ns_err_arr, flux_scaling_arr)


def estimate_ps_sin_dec_discovery_potential_curve(
        ana, rss, fitparam_values, sin_dec_arr, h0_ts_vals_arr,
        h0_ts_quantile=2.7e-3, eps_p=0.0075, mu_min=0, mu_max=20, n_iter=1,
        bkg_kwargs=None, sig_kwargs=None, ppbar=None, **kwargs):
    """Estimates the point-source discovery potential of the given analysis as a
    function of sin(dec). This function places a PointLikeSource source on the
    given declination values and estimates its discovery potential.

    Parameters
    ----------
    ana : Analysis instance
        The instance of Analysis to use for generating trials.
    rss : RandomStateService instance
        The instance of RandomStateService to use for generating random numbers
        from.
    fitparam_values : array_like
        The fit parameter values for which to convert the mean number of signal
        events into a flux.
    sin_dec_arr : (n_sin_dec,)-shaped 1D ndarray
        The ndarray holding the sin(dec) values for which to estimate the
        point-source sensitivity.
    h0_ts_vals_arr : (n_sin_dec,n_iter,n_bkg_trials)-shaped 3D ndarray of float
        The numpy ndarray holding the null-hypothesis ts values for all
        sin(dec) values and iterations.
    h0_ts_quantile : float, optional
        Null-hypothesis test statistic quantile that defines the critical value.
        For a 5sigma discovery potential that value is 5.733e-7. For a 3sigma
        discovery potential this value is 2.7e-3.
        Default is 2.7e-3.
    eps_p : float, optional
        The precision in probability used for the `estimate_discovery_potential`
        function.
        Default is 0.0075.
    mu_min : float, optional
        The minimum value for the mean number of injected signal events as a
        seed for the mu range in which the sensitivity is located.
        Default is 0.
    mu_max : float, optional
        The maximum value for the mean number of injected signal events as a
        seed for the mu range in which the sensitivity is located.
        Default is 20.
    n_iter : int, optional
        Each discovery potential can be estimated several times to be able to
        estimate its variance. This parameter specifies the number of
        iterations to perform for each discovery potential estimation.
        For each iteration the RandomStateService is re-seeded with a seed that
        is incremented by 1.
        Default is 1.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the mean number of signal events, mu.
    ppbar : ProgressBar instance | None
        The optional parent progress bar.

    Additional Keyword Arguments
    ----------------------------
    Any additional keyword arguments are passed to the
    ``estimate_discovery_potential`` function.

    Returns
    -------
    sin_dec_arr : (n_sin_dec,)-shaped 1D ndarray
        The ndarray holding the sin(dec) values for which the discovery
        potential have been estimated.
    mean_ns_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the mean number of signal events corresponding to
        the discovery potential for each sin(dec) value and iteration.
    mean_ns_err_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the estimated error in `mean_ns` for each sin(dec)
        value and iteration.
    flux_scaling_arr : (n_sin_dec,n_iter)-shaped 2D ndarray
        The ndarray holding the scaling factor the reference flux needs to get
        scaled to obtain the flux for the estimated discovery potential.
    """
    mu_min_arr = np.repeat(mu_min, len(sin_dec_arr))
    mu_max_arr = np.repeat(mu_max, len(sin_dec_arr))

    mean_ns_arr = np.empty((len(sin_dec_arr), n_iter))
    mean_ns_err_arr = np.empty((len(sin_dec_arr), n_iter))
    flux_scaling_arr = np.empty((len(sin_dec_arr), n_iter))

    pbar_iter = ProgressBar(n_iter, parent=ppbar).start()
    for iter_idx in range(n_iter):
        pbar = ProgressBar(len(sin_dec_arr), parent=pbar_iter).start()
        for (sin_dec_idx, sin_dec) in enumerate(sin_dec_arr):
            source = PointLikeSource(np.pi, np.arcsin(sin_dec))
            ana.change_source(source)

            h0_ts_vals = h0_ts_vals_arr[sin_dec_idx, iter_idx]

            mu_min = mu_min_arr[sin_dec_idx]
            mu_max = mu_max_arr[sin_dec_idx]

            (mean_ns, mean_ns_err) = estimate_discovery_potential(
                ana, rss, h0_ts_quantile=h0_ts_quantile,
                mu_range=(mu_min, mu_max), eps_p=eps_p,
                h0_ts_vals=h0_ts_vals, bkg_kwargs=bkg_kwargs,
                sig_kwargs=sig_kwargs, ppbar=pbar, **kwargs)

            mean_ns_arr[sin_dec_idx, iter_idx] = mean_ns
            mean_ns_err_arr[sin_dec_idx, iter_idx] = mean_ns_err
            flux_scaling_arr[sin_dec_idx, iter_idx] = ana.calculate_fluxmodel_scaling_factor(
                mean_ns=mean_ns, fitparam_values=np.array(fitparam_values))

            pbar.increment()
        pbar.finish()

        # One iteration is done, update the mu range using the previous results.
        mu_min_arr = np.mean(mean_ns_arr[:, 0:iter_idx+1]*0.8, axis=1)
        mu_max_arr = np.mean(mean_ns_arr[:, 0:iter_idx+1]*1.2, axis=1)

        pbar_iter.increment()

        rss.reseed(rss.seed+1)
    pbar_iter.finish()

    return (sin_dec_arr, mean_ns_arr, mean_ns_err_arr, flux_scaling_arr)



# -*- coding: utf-8 -*-

import abc
import os
import os.path
import shutil
import stat

import numpy as np

from copy import (
    deepcopy,
)

from skyllh.core import (
    display,
)
from skyllh.core.binning import (
    BinningDefinition,
)
from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.datafields import (
    DataFields,
    DataFieldStages as DFS,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.display import (
    ANSIColors,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.py import (
    classname,
    float_cast,
    get_class_of_func,
    issequence,
    issequenceof,
    list_of_cast,
    module_class_method_name,
    str_cast,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
    create_FileLoader,
)
from skyllh.core.timing import (
    TaskTimer,
)


class DatasetOrigin(
    object,
):
    """The DatasetOrigin class provides information about the origin of a
    dataset, so the files of a dataset can be downloaded from the origin.
    """
    def __init__(
            self,
            base_path,
            sub_path,
            transfer_func,
            filename=None,
            host=None,
            port=None,
            username=None,
            password=None,
            post_transfer_func=None,
            **kwargs,
    ):
        """Creates a new instance to define the origin of a dataset.

        Parameters
        ----------
        base_path : str
            The dataset's base directory at the origin.
        sub_path : str
            The dataset's sub directory at the origin.
        transfer_func : callable
            The callable object that should be used to transfer the dataset.
            This function requires the following call signature::

                __call__(origin, file_list, dst_base_path, user=None, password=None)

            where ``origin`` is an instance of DatasetOrigin, ``file_list`` is
            a list of str specifying the files to transfer, ``dst_base_path`` is
            an instance of str specifying the destination base path on the local
            machine, ``user`` is the user name required to connect to the remote
            host, and ``password`` is the password for the user name required to
            connect to the remote host.
        filename : str | None
            If the origin is not a directory but a file, this specifies the
            filename.
        host : str | None
            The name or IP of the remote host.
        port : int | None
            The port number to use when connecting to the remote host.
        username : str | None
            The user name required to connect to the remote host.
        password : str | None
            The password for the user name required to connect to the remote
            host.
        post_transfer_func : callable | None
            The callable object that should be called after the dataset has been
            transferred by the ``transfer_func``function. It can be used to
            extract an archive file.
            This function requires the following call signature::

                __call__(ds, dst_path)

            where ``ds`` is an instance of ``Dataset``, and ``dst_path`` is the
            destination path.
        """
        super().__init__(**kwargs)

        self.base_path = base_path
        self.sub_path = sub_path
        self.transfer_func = transfer_func
        self.filename = filename
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.post_transfer_func = post_transfer_func

    @property
    def base_path(self):
        """The dataset's base directory at the origin.
        """
        return self._base_path

    @base_path.setter
    def base_path(self, obj):
        if not isinstance(obj, str):
            raise TypeError(
                'The base_path property must be an instance of str! '
                f'Its current type is {classname(obj)}!')
        self._base_path = obj

    @property
    def sub_path(self):
        """The dataset's sub directory at the origin.
        """
        return self._sub_path

    @sub_path.setter
    def sub_path(self, obj):
        if not isinstance(obj, str):
            raise TypeError(
                'The sub_path property must be an instance of str! '
                f'Its current type is {classname(obj)}!')
        self._sub_path = obj

    @property
    def root_dir(self):
        """(read-only) The dataset's root directory at the origin, which is
        the combination of ``base_path`` and ``sub_path``.
        """
        return os.path.join(self._base_path, self._sub_path)

    @property
    def transfer_func(self):
        """The callable object that should be used to transfer the dataset.
        """
        return self._transfer_func

    @transfer_func.setter
    def transfer_func(self, obj):
        if not callable(obj):
            raise TypeError(
                'The property transfer_func must be a callable object! '
                f'Its current type is {classname(obj)}!')
        self._transfer_func = obj

    @property
    def filename(self):
        """The file name if the origin is a file instead of a directory.
        """
        return self._filename

    @filename.setter
    def filename(self, obj):
        if obj is not None:
            if not isinstance(obj, str):
                raise TypeError(
                    'The property filename must be None, or an instance of '
                    'str! '
                    f'Its current type is {classname(obj)}!')
        self._filename = obj

    @property
    def is_directory(self):
        """(read-only) Flag if the origin refers to a directory (``True``) or a
        file (``False``).
        """
        return (self._filename is None)

    @property
    def host(self):
        """The name or IP of the remote host.
        """
        return self._host

    @host.setter
    def host(self, obj):
        if obj is not None:
            if not isinstance(obj, str):
                raise TypeError(
                    'The property host must be None, or an instance of str! '
                    f'Its current type is {classname(obj)}!')
        self._host = obj

    @property
    def port(self):
        """The port number to use when connecting to the remote host.
        """
        return self._port

    @port.setter
    def port(self, obj):
        if obj is not None:
            if not isinstance(obj, int):
                raise TypeError(
                    'The property port must be None, or an instance of int! '
                    f'Its current type is {classname(obj)}!')
        self._port = obj

    @property
    def username(self):
        """The user name required to connect to the remote host.
        """
        return self._username

    @username.setter
    def username(self, obj):
        if obj is not None:
            if not isinstance(obj, str):
                raise TypeError(
                    'The property username must be None, or an instance of '
                    'str! '
                    f'Its current type is {classname(obj)}!')
        self._username = obj

    @property
    def password(self):
        """The password for the user name required to connect to the remote
        host.
        """
        return self._password

    @password.setter
    def password(self, obj):
        if obj is not None:
            if not isinstance(obj, str):
                raise TypeError(
                    'The property password must be None, or an instance of '
                    'str! '
                    f'Its current type is {classname(obj)}!')
        self._password = obj

    @property
    def post_transfer_func(self):
        """The callable object that should be called after the dataset has been
        transferred by the ``transfer_func`` callable.
        """
        return self._post_transfer_func

    @post_transfer_func.setter
    def post_transfer_func(self, obj):
        if obj is not None:
            if not callable(obj):
                raise TypeError(
                    'The property post_transfer_func must be a callable '
                    'object! '
                    f'Its current type is {classname(obj)}!')
        self._post_transfer_func = obj

    def __str__(self):
        """Pretty string representation of this class.
        """
        transfer_cls = get_class_of_func(self.transfer_func)

        s = f'{classname(self)} '+'{\n'
        s1 = f'base_path = "{self.base_path}"\n'
        s1 += f'sub_path = "{self.sub_path}"\n'
        if self._filename is not None:
            s1 += f'filename = {self._filename}\n'
        s1 += f'user@host:port = {self.username}@{self.host}:{self.port}\n'
        s1 += 'password = '
        if self.password is not None:
            s1 += 'set\n'
        else:
            s1 += 'not set\n'
        s1 += f'transfer class = {classname(transfer_cls)}\n'
        s += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s1)
        s += '}'

        return s

    def is_locally_available(self):
        """Checks if the dataset origin is available locally by checking if the
        given path exists on the local host.

        Returns
        -------
        check : bool
            ``True`` if the path specified in this dataset origin is an absolute
            path and exists on the local host, ``False`` otherwise.
        """
        root_dir = self.root_dir

        if (
            self.is_directory and
            os.path.abspath(root_dir) == root_dir and
            os.path.exists(root_dir) and
            os.path.isdir(root_dir)
        ):
            return True

        return False


class TemporaryTextFile(
    object,
):
    """This class provides a temporary text file with a given content while
    being within a with statement. Exiting the with statement will remove the
    temporary text file.

    Example:

    .. code::

        with TemporaryTextFile('myfile.txt', 'My file content'):
            # Do something that requires the text file ``myfile.txt``.
        # At this point the text file is removed again.

    """

    def __init__(
            self,
            pathfilename,
            text,
            mode=stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH,
    ):
        self.pathfilename = pathfilename
        self.text = text
        self.mode = mode

    def __enter__(self):
        with open(self.pathfilename, 'w') as fd:
            fd.write(self.text)
        os.chmod(self.pathfilename, self.mode)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.remove(self.pathfilename)


class SystemCommandError(
    Exception,
):
    """This custom exception will be raised when a system command failed.
    """
    pass


class DatasetTransferError(
    Exception,
):
    """This custom exception defines an error that should be raised when the
    actual transfer of the dataset files failed.
    """
    pass


class DatasetTransfer(
    object,
    metaclass=abc.ABCMeta,
):
    """Base class for a dataset transfer mechanism.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def execute_system_command(
            cmd,
            logger,
            success_rcode=0,
    ):
        """Executes the given system command via a ``os.system`` call.

        Parameters
        ----------
        cmd : str
            The system command to execute.
        logger : instance of logging.Logger
            The logger to use for debug messages.
        success_rcode : int | None
            The return code that indicates success of the system command.
            If set to ``None``, no return code checking is performed.

        Raises
        ------
        SystemCommandError
            If the system command did not return ``success_rcode``.
        """
        logger.debug(f'Running command "{cmd}"')
        rcode = os.system(cmd)
        if (success_rcode is not None) and (rcode != success_rcode):
            raise SystemCommandError(
                f'The system command "{cmd}" failed with return code {rcode}!')

    @staticmethod
    def ensure_dst_path(
            dst_path,
    ):
        """Ensures the existence of the given destination path.

        Parameters
        ----------
        dst_path : str
            The destination path.
        """
        if not os.path.isdir(dst_path):
            # Throws if dst_path exists as a file.
            os.makedirs(dst_path)

    @abc.abstractmethod
    def transfer(
            self,
            origin,
            file_list,
            dst_base_path,
            username=None,
            password=None,
    ):
        """This method is supposed to transfer the dataset origin path to the
        given destination path.

        Parameters
        ----------
        origin : instance of DatasetOrigin
            The instance of DatasetOrigin defining the origin of the dataset.
        file_list : list of str
            The list of files, relative to the origin base path, which should be
            transferred.
        dst_base_path : str
            The destination base path into which the dataset files will be
            transferred.
        username : str | None
            The user name required to connect to the remote host.
        password : str | None
            The password for the user name required to connect to the remote
            host.

        Raises
        ------
        DatasetTransferError
            If the actual transfer of the dataset files failed.
        """
        pass

    @staticmethod
    def post_transfer_unzip(
            ds,
            dst_path,
    ):
        """This is a post-transfer function. It will unzip the transferred file
        into the dst_path if the origin path was a zip file.
        """
        if ds.origin.filename is None:
            return

        if not ds.origin.filename.lower().endswith('.zip'):
            return

        cls = get_class_of_func(DatasetTransfer.post_transfer_unzip)
        logger = get_logger(f'{classname(cls)}.post_transfer_unzip')

        # Unzip the dataset file.
        zip_file = os.path.join(dst_path, ds.origin.filename)
        cmd = f'unzip "{zip_file}" -d "{dst_path}"'
        DatasetTransfer.execute_system_command(cmd, logger)

        # Remove the zip file.
        try:
            os.remove(zip_file)
        except Exception as exc:
            logger.warn(str(exc))


class RSYNCDatasetTransfer(
    DatasetTransfer,
):
    def __init__(self, **kwargs):
        super().__init__(
            **kwargs)

    def transfer(  # noqa: C901
            self,
            origin,
            file_list,
            dst_base_path,
            username=None,
            password=None,
    ):
        """Transfers the given dataset to the given destination path using the
        ``rsync`` program.

        Parameters
        ----------
        ds : instance of Dataset
            The instance of Dataset containing the origin property specifying
            the origin of the dataset.
        file_list : list of str
            The list of files, relative to the origin base path, which should be
            transferred.
        dst_base_path : str
            The destination base path into which the dataset files will be
            transferred.
        username : str | None
            The user name required to connect to the remote host.
        password : str | None
            The password for the user name required to connect to the remote
            host.
        """
        cls = get_class_of_func(self.transfer)
        logger = get_logger(f'{classname(cls)}.transfer')

        host = origin.host

        # Make sure the origin and destination base paths end with a directory
        # separator.
        origin_base_path = origin.base_path
        if origin_base_path[-len(os.path.sep):] != os.path.sep:
            origin_base_path += os.path.sep
        if dst_base_path[-len(os.path.sep):] != os.path.sep:
            dst_base_path += os.path.sep

        file_list_pathfilename = os.path.join(
            os.getcwd(),
            f'.{id(origin)}.rsync_file_list.txt')

        # Create file list file content.
        # Skip files which already exists.
        file_list_filecontent = ''
        for file in file_list:
            dst_pathfilename = os.path.join(dst_base_path, file)
            if not os.path.exists(dst_pathfilename):
                file_list_filecontent += f'{file}\n'

        if username is None:
            # No user name is defined.
            with TemporaryTextFile(
                pathfilename=file_list_pathfilename,
                text=file_list_filecontent,
            ):
                cmd = (
                    f'rsync '
                    '-avrRL '
                    '--progress '
                    f'--files-from="{file_list_pathfilename}" '
                    f'{host}:"{origin_base_path}" "{dst_base_path}"'
                )
                try:
                    DatasetTransfer.execute_system_command(cmd, logger)
                except SystemCommandError as err:
                    raise DatasetTransferError(str(err))

        elif password is not None:
            # User and password is defined.
            pwdfile = os.path.join(
                os.getcwd(),
                f'.{id(origin)}.rsync_passwd.txt')

            with TemporaryTextFile(
                pathfilename=pwdfile,
                text=password,
                mode=stat.S_IRUSR,
            ):
                with TemporaryTextFile(
                    pathfilename=file_list_pathfilename,
                    text=file_list_filecontent,
                ):
                    cmd = (
                        f'rsync '
                        '-avrRL '
                        '--progress '
                        f'--password-file "{pwdfile}" '
                        f'--files-from="{file_list_pathfilename}" '
                        f'{username}@{host}:"{origin_base_path}" "{dst_base_path}"'
                    )
                    try:
                        DatasetTransfer.execute_system_command(cmd, logger)
                    except SystemCommandError as err:
                        raise DatasetTransferError(str(err))
        else:
            # Only the user name is defined.
            with TemporaryTextFile(
                pathfilename=file_list_pathfilename,
                text=file_list_filecontent,
            ):
                cmd = (
                    f'rsync '
                    '-avrRL '
                    '--progress '
                    f'--files-from="{file_list_pathfilename}" '
                    f'{username}@{host}:"{origin_base_path}" "{dst_base_path}"'
                )
                try:
                    DatasetTransfer.execute_system_command(cmd, logger)
                except SystemCommandError as err:
                    raise DatasetTransferError(str(err))


class WGETDatasetTransfer(
    DatasetTransfer,
):
    def __init__(self, protocol, **kwargs):
        super().__init__(
            **kwargs)

        self.protocol = protocol

    @property
    def protocol(self):
        """The protocol to use for the transfer.
        """
        return self._protocol

    @protocol.setter
    def protocol(self, obj):
        if not isinstance(obj, str):
            raise TypeError(
                'The property protocol must be an instance of str! '
                f'Its current type is {classname(obj)}!')
        self._protocol = obj

    def transfer(
            self,
            origin,
            file_list,
            dst_base_path,
            username=None,
            password=None,
    ):
        """Transfers the given dataset to the given destination path using the
        ``wget`` program.

        Parameters
        ----------
        origin : instance of DatasetOrigin
            The instance of DatasetOrigin defining the origin of the dataset.
        file_list : list of str
            The list of files relative to the origin's base path, which
            should be transferred.
        dst_base_path : str
            The destination base path into which the dataset will be
            transferred.
        username : str | None
            The user name required to connect to the remote host.
        password : str | None
            The password for the user name required to connect to the remote
            host.
        """
        cls = get_class_of_func(self.transfer)
        logger = get_logger(f'{classname(cls)}.transfer')

        host = origin.host
        port = origin.port

        for file in file_list:
            dst_pathfilename = os.path.join(dst_base_path, file)
            if os.path.exists(dst_pathfilename):
                logger.debug(
                    f'File "{dst_pathfilename}" already exists. Skipping.')
                continue

            path = os.path.join(origin.base_path, file)

            dst_sub_path = os.path.dirname(file)
            if dst_sub_path == '':
                dst_path = dst_base_path
            else:
                dst_path = os.path.join(dst_base_path, dst_sub_path)
            DatasetTransfer.ensure_dst_path(dst_path)

            cmd = 'wget '
            if username is None:
                # No user name is specified.
                pass
            elif password is not None:
                # A user name and password is specified.
                cmd += (
                    f'--user="{username}" '
                    f'--password="{password}" '
                )
            else:
                # Only a user name is specified.
                cmd += (
                    f'--user={username} '
                )
            cmd += f'{self.protocol}://{host}'
            if port is not None:
                cmd += f':{port}'
            if path[0:1] != '/':
                cmd += '/'
            cmd += f'{path} -P {dst_path}'
            try:
                DatasetTransfer.execute_system_command(cmd, logger)
            except SystemCommandError as err:
                raise DatasetTransferError(str(err))


class Dataset(
        HasConfig,
):
    """The Dataset class describes a set of self-consistent experimental and
    simulated detector data. Usually this is for a certain time period, i.e.
    a season.

    Independent data sets of the same kind, e.g. event selection, can be joined
    through a DatasetCollection object.
    """
    @staticmethod
    def get_combined_exp_pathfilenames(
            datasets):
        """Creates the combined list of exp pathfilenames of all the given
        datasets.

        Parameters
        ----------
        datasets : sequence of Dataset
            The sequence of Dataset instances.

        Returns
        -------
        exp_pathfilenames : list
            The combined list of exp pathfilenames.
        """
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The datasets argument must be a sequence of Dataset '
                'instances!')

        exp_pathfilenames = []
        for ds in datasets:
            exp_pathfilenames += ds.exp_pathfilename_list

        return exp_pathfilenames

    @staticmethod
    def get_combined_mc_pathfilenames(
            datasets):
        """Creates the combined list of mc pathfilenames of all the given
        datasets.

        Parameters
        ----------
        datasets : sequence of Dataset
            The sequence of Dataset instances.

        Returns
        -------
        mc_pathfilenames : list
            The combined list of mc pathfilenames.
        """
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The datasets argument must be a sequence of Dataset '
                'instances!')

        mc_pathfilenames = []
        for ds in datasets:
            mc_pathfilenames += ds.mc_pathfilename_list

        return mc_pathfilenames

    @staticmethod
    def get_combined_livetime(
            datasets):
        """Sums the live-time of all the given datasets.

        Parameters
        ----------
        datasets : sequence of Dataset
            The sequence of Dataset instances.

        Returns
        -------
        livetime : float
            The sum of all the individual live-times.
        """
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The datasets argument must be a sequence of Dataset '
                'instances!')

        livetime = np.sum([
            ds.livetime
            for ds in datasets
        ])

        return livetime

    def __init__(
            self,
            name,
            exp_pathfilenames,
            mc_pathfilenames,
            livetime,
            default_sub_path_fmt,
            version,
            verqualifiers=None,
            base_path=None,
            sub_path_fmt=None,
            origin=None,
            **kwargs,
    ):
        """Creates a new dataset object that describes a self-consistent set of
        data.

        Parameters
        ----------
        name : str
            The name of the dataset.
        exp_pathfilenames : str | sequence of str | None
            The file name(s), including paths, of the experimental data file(s).
            This can be None, if a MC-only study is performed.
        mc_pathfilenames : str | sequence of str | None
            The file name(s), including paths, of the monte-carlo data file(s).
            This can be None, if a MC-less analysis is performed.
        livetime : float | None
            The integrated live-time in days of the dataset. It can be None for
            cases where the live-time is retrieved directly from the data files
            upon data loading.
        default_sub_path_fmt : str
            The default format of the sub path of the data set.
            This must be a string that can be formatted via the ``format``
            method of the ``str`` class.
        version : int
            The version number of the dataset. Higher version numbers indicate
            newer datasets.
        verqualifiers : dict | None
            If specified, this dictionary specifies version qualifiers. These
            can be interpreted as subversions of the dataset. The format of the
            dictionary must be 'qualifier (str): version (int)'.
        base_path : str | None
            The user-defined base path of the data set.
            Usually, this is the path of the location of the data directory.
            If set to ``None`` the configured repository base path
            ``Config['repository']['base_path']`` is used.
        sub_path_fmt : str | None
            The user-defined format of the sub path of the data set.
            If set to ``None``, the ``default_sub_path_fmt`` will be used.
        origin : instance of DatasetOrigin | None
            The instance of DatasetOrigin defining the origin of the dataset,
            so the dataset can be transferred automatically to the user's
            device.
        """
        super().__init__(**kwargs)

        self.name = name
        self.exp_pathfilename_list = exp_pathfilenames
        self.mc_pathfilename_list = mc_pathfilenames
        self.livetime = livetime
        self.default_sub_path_fmt = default_sub_path_fmt
        self.version = version
        self.verqualifiers = verqualifiers
        self.base_path = base_path
        self.sub_path_fmt = sub_path_fmt
        self.origin = origin

        self.description = ''

        self._datafields = dict()

        self._exp_field_name_renaming_dict = dict()
        self._mc_field_name_renaming_dict = dict()

        self._data_preparation_functions = list()
        self._binning_definitions = dict()
        self._aux_data_definitions = dict()
        self._aux_data = dict()

    @property
    def name(self):
        """The name of the dataset. This must be an unique identifier among
        all the different datasets.
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def description(self):
        """The (longer) description of the dataset.
        """
        return self._description

    @description.setter
    def description(self, description):
        if not isinstance(description, str):
            raise TypeError(
                'The description of the dataset must be of type str!')
        self._description = description

    @property
    def datafields(self):
        """The dictionary holding the names and stages of required data fields
        specific for this dataset.
        """
        return self._datafields

    @datafields.setter
    def datafields(self, fields):
        if not isinstance(fields, dict):
            raise TypeError(
                'The datafields property must be a dictionary! '
                f'Its current type is "{classname(fields)}"!')
        self._datafields = fields

    @property
    def exp_pathfilename_list(self):
        """The list of file names of the data files that store the experimental
        data for this dataset.
        If a file name is given with a relative path, it will be relative to the
        root_dir property of this Dataset instance.
        """
        return self._exp_pathfilename_list

    @exp_pathfilename_list.setter
    def exp_pathfilename_list(self, pathfilenames):
        if pathfilenames is None:
            pathfilenames = []
        if isinstance(pathfilenames, str):
            pathfilenames = [pathfilenames]
        if not issequenceof(pathfilenames, str):
            raise TypeError(
                'The exp_pathfilename_list property must be of type str or a '
                'sequence of str!')
        self._exp_pathfilename_list = list(pathfilenames)

    @property
    def exp_abs_pathfilename_list(self):
        """(read-only) The list of absolute path file names of the experimental
        data files.
        """
        return self.get_abs_pathfilename_list(self._exp_pathfilename_list)

    @property
    def mc_pathfilename_list(self):
        """The list of file names of the data files that store the monte-carlo
        data for this dataset.
        If a file name is given with a relative path, it will be relative to the
        root_dir property of this Dataset instance.
        """
        return self._mc_pathfilename_list

    @mc_pathfilename_list.setter
    def mc_pathfilename_list(self, pathfilenames):
        if pathfilenames is None:
            pathfilenames = []
        if isinstance(pathfilenames, str):
            pathfilenames = [pathfilenames]
        if not issequenceof(pathfilenames, str):
            raise TypeError(
                'The mc_pathfilename_list property must be of type str or a '
                'sequence of str!')
        self._mc_pathfilename_list = list(pathfilenames)

    @property
    def mc_abs_pathfilename_list(self):
        """(read-only) The list of absolute path file names of the monte-carlo
        data files.
        """
        return self.get_abs_pathfilename_list(self._mc_pathfilename_list)

    @property
    def livetime(self):
        """The integrated live-time in days of the dataset. This can be None in
        cases where the livetime is retrieved directly from the data files.
        """
        return self._lifetime

    @livetime.setter
    def livetime(self, lt):
        if lt is not None:
            lt = float_cast(
                lt,
                'The lifetime property of the dataset must be cast-able to '
                'type float!')
        self._lifetime = lt

    @property
    def version(self):
        """The main version (int) of the dataset.
        """
        return self._version

    @version.setter
    def version(self, version):
        if not isinstance(version, int):
            raise TypeError(
                'The version of the dataset must be of type int!')
        self._version = version

    @property
    def verqualifiers(self):
        """The dictionary holding the version qualifiers, i.e. sub-version
        qualifiers. If set to None, an empty dictionary will be used.
        The dictionary must have the type form of str:int.
        """
        return self._verqualifiers

    @verqualifiers.setter
    def verqualifiers(self, verqualifiers):
        if verqualifiers is None:
            verqualifiers = dict()
        if not isinstance(verqualifiers, dict):
            raise TypeError('The version qualifiers must be of type dict!')
        # Check if the dictionary has format str:int.
        for (q, v) in verqualifiers.items():
            if not isinstance(q, str):
                raise TypeError(
                    f'The version qualifier "{q}" must be of type str!')
            if not isinstance(v, int):
                raise TypeError(
                    f'The version for the qualifier "{q}" must be of type int!')
        # We need to take a deep copy in order to make sure that two datasets
        # don't share the same version qualifier dictionary.
        self._verqualifiers = deepcopy(verqualifiers)

    @property
    def base_path(self):
        """The base path of the data set. This can be ``None``.
        """
        return self._base_path

    @base_path.setter
    def base_path(self, path):
        if path is not None:
            path = str_cast(
                path,
                'The base_path property must be cast-able to type str!')
            if not os.path.isabs(path):
                raise ValueError(
                    'The base_path property must be an absolute path!')
        self._base_path = path

    @property
    def default_sub_path_fmt(self):
        """The default format of the sub path of the data set. This must be a
        string that can be formatted via the ``format`` method of the ``str``
        class.
        """
        return self._default_sub_path_fmt

    @default_sub_path_fmt.setter
    def default_sub_path_fmt(self, fmt):
        fmt = str_cast(
            fmt,
            'The default_sub_path_fmt property must be cast-able to type str!')
        self._default_sub_path_fmt = fmt

    @property
    def sub_path_fmt(self):
        """The format of the sub path of the data set. This must be a string
        that can be formatted via the ``format`` method of the ``str`` class.
        If set to ``None``, this property will return the
        ``default_sub_path_fmt`` property.
        """
        if self._sub_path_fmt is None:
            return self._default_sub_path_fmt
        return self._sub_path_fmt

    @sub_path_fmt.setter
    def sub_path_fmt(self, fmt):
        if fmt is not None:
            fmt = str_cast(
                fmt,
                'The sub_path_fmt property must be None, or cast-able to type '
                'str!')
        self._sub_path_fmt = fmt

    @property
    def origin(self):
        """The instance of DatasetOrigin defining the origin of the dataset.
        This can be ``None`` if the dataset has no origin defined.
        """
        return self._origin

    @origin.setter
    def origin(self, obj):
        if obj is not None:
            if not isinstance(obj, DatasetOrigin):
                raise TypeError(
                    'The origin property must be None, or an instance of '
                    'DatasetOrigin! '
                    f'Its current type is {classname(obj)}!')
        self._origin = obj

    @property
    def root_dir(self):
        """(read-only) The root directory to use when data files are specified
        with relative paths. It is constructed from the ``base_path`` and the
        ``sub_path_fmt`` properties via the ``generate_data_file_root_dir``
        function.
        """
        return generate_data_file_root_dir(
            default_base_path=self._cfg['repository']['base_path'],
            default_sub_path_fmt=self._default_sub_path_fmt,
            version=self._version,
            verqualifiers=self._verqualifiers,
            base_path=self._base_path,
            sub_path_fmt=self._sub_path_fmt)

    @property
    def exp_field_name_renaming_dict(self):
        """The dictionary specifying the field names of the experimental data
        which need to get renamed just after loading the data. The dictionary
        values are the new names.
        """
        return self._exp_field_name_renaming_dict

    @exp_field_name_renaming_dict.setter
    def exp_field_name_renaming_dict(self, d):
        if not isinstance(d, dict):
            raise TypeError(
                'The exp_field_name_renaming_dict property must be an instance '
                'of dict!')
        self._exp_field_name_renaming_dict = d

    @property
    def mc_field_name_renaming_dict(self):
        """The dictionary specifying the field names of the monte-carlo data
        which need to get renamed just after loading the data. The dictionary
        values are the new names.
        """
        return self._mc_field_name_renaming_dict

    @mc_field_name_renaming_dict.setter
    def mc_field_name_renaming_dict(self, d):
        if not isinstance(d, dict):
            raise TypeError(
                'The mc_field_name_renaming_dict property must be an instance '
                'of dict!')
        self._mc_field_name_renaming_dict = d

    @property
    def exists(self):
        """(read-only) Flag if all the data files of this dataset exists. It is
        ``True`` if all data files exist and ``False`` otherwise.
        """
        file_list = self.create_file_list()
        abs_file_list = self.get_abs_pathfilename_list(file_list)
        for abs_file in abs_file_list:
            if not os.path.exists(abs_file):
                return False
        return True

    @property
    def version_str(self):
        """The version string of the dataset. This combines all the version
        information about the dataset.
        """
        s = f'{self._version:03d}'
        for (q, v) in self._verqualifiers.items():
            s += f'{q}{v:02d}'
        return s

    @property
    def data_preparation_functions(self):
        """The list of callback functions that will be called to prepare the
        data (experimental and monte-carlo).
        """
        return self._data_preparation_functions

    def _gen_datafile_pathfilename_entry(self, pathfilename):
        """Generates a string containing the given pathfilename and if it exists
        (FOUND) or not (NOT FOUND).

        Parameters
        ----------
        pathfilename : str
            The fully qualified path and filename of the file.

        Returns
        -------
        s : str
            The generated string.
        """
        if os.path.exists(pathfilename):
            s = '['+ANSIColors.OKGREEN+'FOUND'+ANSIColors.ENDC+']'
        else:
            s = '['+ANSIColors.FAIL+'NOT FOUND'+ANSIColors.ENDC+']'
        s += ' ' + pathfilename
        return s

    def __gt__(self, ds):
        """Implementation to support the operation ``b = self > ds``, where
        ``self`` is this Dataset object and ``ds`` an other Dataset object.
        The comparison is done based on the version information of both
        datasets. Larger version numbers for equal version qualifiers indicate
        newer (greater) datasets.

        The two datasets must be of the same kind, i.e. have the same name, in
        order to make the version comparison senseful.

        Returns
        -------
        bool
            True, if this dataset is newer than the reference dataset.
            False, if this dataset is as new or older than the reference
            dataset.
        """
        # Datasets of different names cannot be compared usefully.
        if self._name != ds._name:
            return False

        # Larger main version numbers indicate newer datasets.
        if self._version > ds._version:
            return True

        # Look for version qualifiers that make this dataset older than the
        # reference dataset.
        qs1 = self._verqualifiers.keys()
        qs2 = ds._verqualifiers.keys()

        # If a qualifier of self is also specified for ds, the version number
        # of the self qualifier must be larger than the version number of the ds
        # qualifier, in order to consider self as newer dataset.
        # If a qualifier is present in self but not in ds, self is considered
        # newer.
        for q in qs1:
            if q in qs2 and qs1[q] <= qs2[q]:
                return False
        # If there is a qualifier in ds but not in self, self is considered
        # older.
        for q in qs2:
            if q not in qs1:
                return False

        return True

    def __str__(self):  # noqa: C901
        """Implementation of the pretty string representation of the Dataset
        object.
        """
        s = f'Dataset "{self.name}": v{self.version_str}\n'

        s1 = ''

        if self.livetime is None:
            s1 += '{ livetime = UNDEFINED }'
        else:
            s1 += '{ 'f'livetime = {self.livetime:.3f} days'' }'
        s1 += '\n'

        if self.description != '':
            s1 += 'Description:\n' + self.description + '\n'

        s1 += 'Experimental data:\n'
        s2 = ''
        for (idx, pathfilename) in enumerate(self.exp_abs_pathfilename_list):
            if idx > 0:
                s2 += '\n'
            s2 += self._gen_datafile_pathfilename_entry(pathfilename)
        s1 += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s2)
        s1 += '\n'

        s1 += 'MC data:\n'
        s2 = ''
        for (idx, pathfilename) in enumerate(self.mc_abs_pathfilename_list):
            if idx > 0:
                s2 += '\n'
            s2 += self._gen_datafile_pathfilename_entry(pathfilename)
        s1 += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s2)
        s1 += '\n'

        if len(self._aux_data_definitions) > 0:
            s1 += 'Auxiliary data:\n'
            s2 = ''
            for (idx, (name, pathfilename_list)) in enumerate(
                    self._aux_data_definitions.items()):
                if idx > 0:
                    s2 += '\n'

                s2 += name+':'
                s3 = ''
                pathfilename_list = self.get_abs_pathfilename_list(
                    pathfilename_list)
                for pathfilename in pathfilename_list:
                    s3 += '\n' + self._gen_datafile_pathfilename_entry(pathfilename)
                s2 += display.add_leading_text_line_padding(
                    display.INDENTATION_WIDTH, s3)
            s1 += display.add_leading_text_line_padding(
                display.INDENTATION_WIDTH, s2)

        s += display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s1)

        return s

    def remove_data(self):
        """Removes the data of this dataset by removing the dataset's root
        directory and everything in it. If the root directory is a symbolic
        link, only this link will be removed.

        Raises
        ------
        RuntimeError
            If the dataset's root directory is neither a symlink nor a
            directory.
        """
        root_dir = self.root_dir

        if os.path.islink(root_dir):
            os.remove(root_dir)
            return

        if os.path.isdir(root_dir):
            shutil.rmtree(root_dir)
            return

        raise RuntimeError(
            f'The root directory "{root_dir}" of dataset {self.name} is '
            'neither a symlink nor a directory!')

    def get_abs_pathfilename_list(
            self,
            pathfilename_list,
    ):
        """Returns a list where each entry of the given pathfilename_list is
        an absolute path. Relative paths will be prefixed with the root_dir
        property of this Dataset instance.

        Parameters
        ----------
        pathfilename_list : sequence of str
            The sequence of file names, either with relative, or absolute paths.

        Returns
        -------
        abs_pathfilename_list : list of str
            The list of file names with absolute paths.
        """
        root_dir = self.root_dir

        abs_pathfilename_list = []
        for pathfilename in pathfilename_list:
            if os.path.isabs(pathfilename):
                abs_pathfilename_list.append(
                    pathfilename)
            else:
                abs_pathfilename_list.append(
                    os.path.join(root_dir, pathfilename))

        return abs_pathfilename_list

    def get_missing_files(
            self,
    ):
        """Determines which files of the dataset are missing and returns the
        list of files.

        Returns
        -------
        missing_files : list of str
            The list of files that are missing. The files are relative to the
            dataset's root directory.
        """
        file_list = self.create_file_list()
        abs_file_list = self.get_abs_pathfilename_list(file_list)

        missing_files = [
            file
            for (file, abs_file) in zip(file_list, abs_file_list)
            if not os.path.exists(abs_file)
        ]

        return missing_files

    def update_version_qualifiers(
            self,
            verqualifiers,
    ):
        """Updates the version qualifiers of the dataset. The update can only
        be done by increasing the version qualifier integer or by adding new
        version qualifiers.

        Parameters
        ----------
        verqualifiers : dict
            The dictionary with the new version qualifiers.

        Raises
        ------
        ValueError
            If the integer number of an existing version qualifier is not larger
            than the old one.
        """
        got_new_verqualifiers = False
        verqualifiers_keys = verqualifiers.keys()
        self_verqualifiers_keys = self._verqualifiers.keys()
        if len(verqualifiers_keys) > len(self_verqualifiers_keys):
            # New version qualifiers must be a subset of the old version
            # qualifiers.
            for q in self_verqualifiers_keys:
                if q not in verqualifiers_keys:
                    raise ValueError(
                        f'The version qualifier {q} has been dropped!')
            got_new_verqualifiers = True

        existing_verqualifiers_incremented = False
        for q in verqualifiers:
            if (q in self._verqualifiers) and\
               (verqualifiers[q] > self._verqualifiers[q]):
                existing_verqualifiers_incremented = True
            self._verqualifiers[q] = verqualifiers[q]

        if not (got_new_verqualifiers or existing_verqualifiers_incremented):
            raise ValueError(
                'Version qualifier values did not increment and no new version '
                'qualifiers were added!')

    def create_file_list(
            self,
    ):
        """Creates the list of files that are linked to this dataset.
        The file paths are relative to the dataset's root directory.

        Returns
        -------
        file_list : list of str
            The list of files of this dataset.
        """
        file_list = (
            self._exp_pathfilename_list +
            self._mc_pathfilename_list
        )

        for aux_pathfilename_list in self._aux_data_definitions.values():
            file_list += aux_pathfilename_list

        return file_list

    def make_data_available(  # noqa: C901
            self,
            username=None,
            password=None,
    ):
        """Makes the data of the dataset available.
        If the root directory of the dataset does not exist locally, the dataset
        is transferred from its origin to the local host. If the origin is
        already available locally, only a symlink is created to the origin path.

        Parameters
        ----------
        username : str | None
            The user name required to connect to the remote host of the origin.
            If set to ``None``, the
        password : str | None
            The password of the user name required to connect to the remote host
            of the origin.

        Returns
        -------
        success : bool
            ``True`` if the data was made available successfully, ``False``
            otherwise.
        """
        logger = get_logger(
            module_class_method_name(self, 'make_data_available')
        )

        if len(self.get_missing_files()) == 0:
            logger.debug(
                f'All files of dataset "{self.name}" already exist. '
                'Nothing to download.')
            return True

        if self.origin is None:
            logger.warn(
                f'No origin defined for dataset "{self.name}"! '
                'Cannot download dataset!')
            return False

        # Check if the dataset origin is locally available. In that case we
        # just create a symlink.
        if self.origin.is_locally_available():
            root_dir = self.root_dir

            # Check if the symlink to the root directory already exists.
            if os.path.isdir(root_dir):
                return True

            # Make sure all directories leading to the symlink exist.
            dirname = os.path.dirname(root_dir)
            if dirname != '':
                os.makedirs(dirname, exist_ok=True)

            cmd = f'ln -s "{self.origin.root_dir}" "{root_dir}"'
            DatasetTransfer.execute_system_command(cmd, logger)
            return True

        if self._cfg['repository']['download_from_origin'] is False:
            logger.warn(
                f'The data of dataset "{self.name}" is locally not available '
                'and the download from the origin is disabled through the '
                'configuration!')
            return False

        if username is None:
            username = self.origin.username
        if password is None:
            password = self.origin.password

        base_path = generate_base_path(
            default_base_path=self._cfg['repository']['base_path'],
            base_path=self._base_path)

        logger.debug(
            f'Downloading dataset "{self.name}" from origin into base path '
            f'"{base_path}". username="{username}".')

        # Check if the origin is a directory. If not we just transfer that one
        # file.
        if self.origin.is_directory:
            file_list = [
                os.path.join(self.origin.sub_path, pathfilename)
                for pathfilename in self.create_file_list()
            ]
        else:
            file_list = [
                os.path.join(self.origin.sub_path, self.origin.filename)
            ]

        self.origin.transfer_func(
            origin=self.origin,
            file_list=file_list,
            dst_base_path=base_path,
            username=username,
            password=password)

        if self.origin.post_transfer_func is not None:
            self.origin.post_transfer_func(
                ds=self,
                dst_path=base_path)

        return True

    def load_data(
            self,
            livetime=None,
            keep_fields=None,
            dtc_dict=None,
            dtc_except_fields=None,
            efficiency_mode=None,
            tl=None,
    ):
        """Loads the data, which is described by the dataset.

        .. note:

            This does not call the ``prepare_data`` method! It only loads the
            data as the method names says.

        Parameters
        ----------
        livetime : instance of Livetime | float | None
            If not None, uses this livetime (if float, livetime in days) for the
            DatasetData instance, otherwise uses the Dataset livetime property
            value for the DatasetData instance.
        keep_fields : list of str | None
            The list of user-defined data fields that should get loaded and kept
            in addition to the analysis required data fields.
        dtc_dict : dict | None
            This dictionary defines how data fields of specific data types (key)
            should get converted into other data types (value).
            This can be used to use less memory. If set to None, no data
            conversion is performed.
        dtc_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                ``'memory'``
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                ``'time'``
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.
        tl : instance of TimeLord | None
            The TimeLord instance to use to time the data loading procedure.

        Returns
        -------
        data : instance of DatasetData
            A instance of DatasetData holding the experimental and monte-carlo
            data.
        """
        def _conv_new2orig_field_names(
                new_field_names,
                orig2new_renaming_dict,
        ):
            """Converts the given ``new_field_names`` into their original name
            given the original-to-new field name renaming dictionary.
            """
            if new_field_names is None:
                return None

            new2orig_renaming_dict = {
                v: k
                for (k, v) in orig2new_renaming_dict.items()
            }

            orig_field_names = [
                new2orig_renaming_dict.get(new_field_name, new_field_name)
                for new_field_name in new_field_names
            ]

            return orig_field_names

        if self._cfg['repository']['download_from_origin'] is True:
            self.make_data_available()

        if keep_fields is None:
            keep_fields = []

        datafields = {**self._cfg['datafields'], **self._datafields}

        # Load the experimental data if there is any.
        if len(self._exp_pathfilename_list) > 0:
            with TaskTimer(tl, 'Loading exp data from disk.'):
                fileloader_exp = create_FileLoader(
                    self.exp_abs_pathfilename_list)
                # Create the list of field names that should get kept.
                keep_fields_exp = list(set(
                    _conv_new2orig_field_names(
                        DataFields.get_joint_names(
                            datafields=datafields,
                            stages=(
                                DFS.DATAPREPARATION_EXP |
                                DFS.ANALYSIS_EXP
                            )
                        ) +
                        keep_fields,
                        self._exp_field_name_renaming_dict
                    )
                ))

                data_exp = fileloader_exp.load_data(
                    keep_fields=keep_fields_exp,
                    dtype_conversions=dtc_dict,
                    dtype_conversion_except_fields=_conv_new2orig_field_names(
                        dtc_except_fields,
                        self._exp_field_name_renaming_dict),
                    efficiency_mode=efficiency_mode)
                data_exp.rename_fields(self._exp_field_name_renaming_dict)
        else:
            data_exp = None

        # Load the monte-carlo data if there is any.
        if len(self._mc_pathfilename_list) > 0:
            with TaskTimer(tl, 'Loading mc data from disk.'):
                fileloader_mc = create_FileLoader(
                    self.mc_abs_pathfilename_list)
                # Determine `keep_fields_mc` for the generic case, where MC
                # field names are an union of exp and mc field names.
                # But the renaming dictionary can differ for exp and MC fields.
                keep_fields_mc = list(set(
                    _conv_new2orig_field_names(
                        DataFields.get_joint_names(
                            datafields=datafields,
                            stages=(
                                DFS.DATAPREPARATION_EXP |
                                DFS.ANALYSIS_EXP
                            )
                        ) +
                        keep_fields,
                        self._exp_field_name_renaming_dict
                    ) +
                    _conv_new2orig_field_names(
                        DataFields.get_joint_names(
                            datafields=datafields,
                            stages=(
                                DFS.DATAPREPARATION_EXP |
                                DFS.ANALYSIS_EXP |
                                DFS.DATAPREPARATION_MC |
                                DFS.ANALYSIS_MC
                            )
                        ) +
                        keep_fields,
                        self._mc_field_name_renaming_dict
                    )
                ))
                data_mc = fileloader_mc.load_data(
                    keep_fields=keep_fields_mc,
                    dtype_conversions=dtc_dict,
                    dtype_conversion_except_fields=_conv_new2orig_field_names(
                        dtc_except_fields,
                        self._mc_field_name_renaming_dict),
                    efficiency_mode=efficiency_mode)
                data_mc.rename_fields(self._mc_field_name_renaming_dict)
        else:
            data_mc = None

        if livetime is None:
            livetime = self.livetime

        data = DatasetData(
            data_exp=data_exp,
            data_mc=data_mc,
            livetime=livetime)

        return data

    def load_aux_data(
            self,
            name,
            tl=None,
    ):
        """Loads the auxiliary data for the given auxiliary data definition.

        Parameters
        ----------
        name : str
            The name of the auxiliary data.
        tl : instance of TimeLord | None
            The TimeLord instance to use to time the data loading procedure.

        Returns
        -------
        data : unspecified
            The loaded auxiliary data.
        """
        name = str_cast(
            name,
            'The name argument must be cast-able to type str!')

        # Check if the data was defined in memory.
        if name in self._aux_data:
            with TaskTimer(tl, f'Loaded aux data "{name}" from memory.'):
                data = self._aux_data[name]
            return data

        if name not in self._aux_data_definitions:
            raise KeyError(
                f'The auxiliary data named "{name}" does not exist!')

        aux_pathfilename_list = self._aux_data_definitions[name]
        with TaskTimer(tl, f'Loaded aux data "{name}" from disk.'):
            fileloader_aux = create_FileLoader(self.get_abs_pathfilename_list(
                aux_pathfilename_list))
            data = fileloader_aux.load_data()

        return data

    def add_data_preparation(
            self,
            func,
    ):
        """Adds the given data preparation function to the dataset.

        Parameters
        ----------
        func : callable
            The object with call signature __call__(data) that will prepare
            the data after it was loaded. The argument 'data' is a DatasetData
            instance holding the experimental and monte-carlo data. The function
            must alter the properties of the DatasetData instance.

        """
        if not callable(func):
            raise TypeError(
                'The argument "func" must be a callable object with call '
                'signature __call__(data)!')
        self._data_preparation_functions.append(func)

    def remove_data_preparation(
            self,
            key=-1,
    ):
        """Removes a data preparation function from the dataset.

        Parameters
        ----------
        key : str, int, optional
            The name or the index of the data preparation function that should
            be removed. Default value is ``-1``, i.e. the last added function.

        Raises
        ------
        TypeError
            If the type of the key argument is invalid.
        IndexError
            If the given key is out of range.
        KeyError
            If the data preparation function cannot be found.
        """
        if isinstance(key, int):
            n = len(self._data_preparation_functions)
            if (key < -n) or (key >= n):
                raise IndexError(
                    f'The given index ({key}) for the data preparation '
                    f'function is out of range ({-n},{n-1})!')
            del self._data_preparation_functions[key]
            return
        elif isinstance(key, str):
            for (i, func) in enumerate(self._data_preparation_functions):
                if func.__name__ == key:
                    del self._data_preparation_functions[i]
                    return
            raise KeyError(
                f'The data preparation function "{key}" was not found in the '
                f'dataset "{self._name}"!')

        TypeError(
            'The key argument must be an instance of int or str!')

    def prepare_data(
            self,
            data,
            tl=None,
    ):
        """Prepares the data by calling the data preparation callback functions
        of this dataset.

        Parameters
        ----------
        data : instance of DatasetData
            The instance of DatasetData holding the data.
        tl : instance of TimeLord | None
            The instance TimeLord that should be used to time the data
            preparation.
        """
        for data_prep_func in self._data_preparation_functions:
            with TaskTimer(
                    tl,
                    f'Preparing data of dataset "{self.name}" by '
                    f'"{data_prep_func.__name__}".'):
                data_prep_func(data)

    def load_and_prepare_data(
            self,
            livetime=None,
            keep_fields=None,
            dtc_dict=None,
            dtc_except_fields=None,
            efficiency_mode=None,
            tl=None,
    ):
        """Loads and prepares the experimental and monte-carlo data of this
        dataset by calling its ``load_data`` and ``prepare_data`` methods.
        After loading the data it drops all unnecessary data fields if they are
        not listed in ``keep_fields``.
        In the end it asserts the data format of the experimental and
        monte-carlo data.

        Parameters
        ----------
        livetime : float | None
            The user-defined livetime in days of the data set. If not set to
            None, livetime information from the data set will get ignored and
            this value of the livetime will be used.
        keep_fields : sequence of str | None
            The list of additional data fields that should get kept.
            By default only the required data fields are kept.
        dtc_dict : dict | None
            This dictionary defines how data fields of specific data types (key)
            should get converted into other data types (value).
            This can be used to use less memory. If set to None, no data
            conversion is performed.
        dtc_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                ``'memory'``
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                ``'time'``
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time the data
            loading and preparation.

        Returns
        -------
        data : instance of DatasetData
            The instance of DatasetData holding the experimental and monte-carlo
            data.
        """
        if keep_fields is None:
            keep_fields = list()
        elif not issequenceof(keep_fields, str):
            raise TypeError(
                'The keep_fields argument must be None, or a sequence of str!')
        keep_fields = list(keep_fields)

        data = self.load_data(
            keep_fields=keep_fields,
            livetime=livetime,
            dtc_dict=dtc_dict,
            dtc_except_fields=dtc_except_fields,
            efficiency_mode=efficiency_mode,
            tl=tl)

        self.prepare_data(data, tl=tl)

        # Drop non-required data fields.
        if data.exp is not None:
            with TaskTimer(tl, 'Cleaning exp data.'):
                keep_fields_exp = (
                    DataFields.get_joint_names(
                        datafields=self._cfg['datafields'],
                        stages=(
                            DFS.ANALYSIS_EXP
                        )
                    ) +
                    keep_fields
                )
                data.exp.tidy_up(keep_fields=keep_fields_exp)

        if data.mc is not None:
            with TaskTimer(tl, 'Cleaning MC data.'):
                keep_fields_mc = (
                    DataFields.get_joint_names(
                        datafields=self._cfg['datafields'],
                        stages=(
                            DFS.ANALYSIS_EXP |
                            DFS.ANALYSIS_MC
                        )
                    ) +
                    keep_fields
                )
                data.mc.tidy_up(keep_fields=keep_fields_mc)

        with TaskTimer(tl, 'Asserting data format.'):
            assert_data_format(self, data)

        return data

    def add_binning_definition(
            self,
            binning,
    ):
        """Adds a binning setting to this dataset.

        Parameters
        ----------
        binning : BinningDefinition
            The BinningDefinition object holding the binning information.
        """
        if not isinstance(binning, BinningDefinition):
            raise TypeError(
                'The "binning" argument must be of type BinningDefinition!')
        if binning.name in self._binning_definitions:
            raise KeyError(
                f'The binning definition "{binning.name}" is already defined '
                f'for dataset "{self._name}"!')

        self._binning_definitions[binning.name] = binning

    def get_binning_definition(
            self,
            name,
    ):
        """Gets the BinningDefinition object for the given binning name.

        Parameters
        ----------
        name : str
            The name of the binning definition.

        Returns
        -------
        binning_definition : instance of BinningDefinition
            The requested instance of BinningDefinition.
        """
        if name not in self._binning_definitions:
            raise KeyError(
                f'The given binning name "{name}" has not been added to the '
                'dataset yet!')
        return self._binning_definitions[name]

    def remove_binning_definition(
            self,
            name,
    ):
        """Removes the BinningDefinition object from the dataset.

        Parameters
        ----------
        name : str
            The name of the binning definition.

        """
        if name not in self._binning_definitions:
            raise KeyError(
                f'The given binning name "{name}" does not exist in the '
                f'dataset "{self.name}", nothing to remove!')

        self._binning_definitions.pop(name)

    def has_binning_definition(
            self,
            name,
    ):
        """Checks if the dataset has a defined binning definition with the given
        name.

        Parameters
        ----------
        name : str
            The name of the binning definition.

        Returns
        -------
        check : bool
            True if the binning definition exists, False otherwise.
        """
        if name in self._binning_definitions:
            return True
        return False

    def define_binning(
            self,
            name,
            binedges,
    ):
        """Defines a binning for ``name``, and adds it as binning definition.

        Parameters
        ----------
        name : str
            The name of the binning setting.
        binedges : sequence
            The sequence of the bin edges, which should be used for this binning
            definition.

        Returns
        -------
        binning : instance of BinningDefinition
            The instance of BinningDefinition which was created and added to
            this dataset.
        """
        binning = BinningDefinition(name, binedges)
        self.add_binning_definition(binning)
        return binning

    def replace_binning_definition(self, binning):
        """Replaces an already defined binning definition of this dataset by
        the given binning definition.

        Parameters
        ----------
        binning : instance of BinningDefinition
            The instance of BinningDefinition that will replace the dataset's
            BinningDefinition instance of the same name.
        """
        if not isinstance(binning, BinningDefinition):
            raise TypeError(
                'The "binning" argument must be of type BinningDefinition!')
        if binning.name not in self._binning_definitions:
            raise KeyError(
                f'The given binning definition "{binning.name}" has not been '
                'added to the dataset yet!')

        self._binning_definitions[binning.name] = binning

    def add_aux_data_definition(
            self,
            name,
            pathfilenames,
    ):
        """Adds the given data files as auxiliary data definition to the
        dataset.

        Parameters
        ----------
        name : str
            The name of the auxiliary data definition. The name is used as
            identifier for the data within SkyLLH.
        pathfilenames : str | sequence of str
            The file name(s) (including paths) of the data file(s).
        """
        name = str_cast(
            name,
            'The name argument must be cast-able to type str! '
            f'Its current type is {classname(name)}.')

        pathfilenames = list_of_cast(
            str,
            pathfilenames,
            'The pathfilenames argument must be of type str or a sequence '
            f'of str! Its current type is {classname(pathfilenames)}.')

        if name in self._aux_data_definitions:
            raise KeyError(
                f'The auxiliary data definition "{name}" is already defined '
                f'for dataset "{self.name}"!')

        self._aux_data_definitions[name] = pathfilenames

    def get_aux_data_definition(
            self,
            name,
    ):
        """Returns the auxiliary data definition from the dataset.

        Parameters
        ----------
        name : str
            The name of the auxiliary data definition.

        Raises
        ------
        KeyError
            If auxiliary data with the given name does not exist.

        Returns
        -------
        aux_data_definition : list of str
            The locations (pathfilenames) of the files defined in the auxiliary
            data as auxiliary data definition.
        """
        if name not in self._aux_data_definitions:
            raise KeyError(
                f'The auxiliary data definition "{name}" does not exist in '
                f'dataset "{self.name}"!')

        return self._aux_data_definitions[name]

    def set_aux_data_definition(
            self,
            name,
            pathfilenames,
    ):
        """Sets the files of the auxiliary data definition, which has the given
        name.

        Parameters
        ----------
        name : str
            The name of the auxiliary data definition.
        pathfilenames : str | sequence of str
            The file name(s) (including paths) of the data file(s).
        """
        name = str_cast(
            name,
            'The name argument must be cast-able to type str! '
            f'Its current type is {classname(name)}.')

        pathfilenames = list_of_cast(
            str,
            pathfilenames,
            'The pathfilenames argument must be of type str or a sequence '
            f'of str! Its current type is {classname(pathfilenames)}.')

        if name not in self._aux_data_definitions:
            raise KeyError(
                f'The auxiliary data definition "{name}" is not defined '
                f'for dataset "{self.name}"! Use add_aux_data_definition '
                'instead!')

        self._aux_data_definitions[name] = pathfilenames

    def remove_aux_data_definition(
            self,
            name,
    ):
        """Removes the auxiliary data definition from the dataset.

        Parameters
        ----------
        name : str
            The name of the data definition that should get removed.
        """
        if name not in self._aux_data_definitions:
            raise KeyError(
                f'The auxiliary data definition "{name}" does not exist in '
                f'dataset "{self.name}", nothing to remove!')

        self._aux_data_definitions.pop(name)

    def remove_aux_data_definitions(
            self,
            names,
    ):
        """Removes the auxiliary data definition from the dataset.

        Parameters
        ----------
        names : sequence of str
            The names of the data definitions that should get removed.
        """
        for name in names:
            self.remove_aux_data_definition(
                name=name)

    def add_aux_data(
            self,
            name,
            data,
    ):
        """Adds the given data as auxiliary data to this data set.

        Parameters
        ----------
        name : str
            The name under which the auxiliary data will be stored.
        data : unspecified
            The data that should get stored. This can be of any data type.

        Raises
        ------
        KeyError
            If auxiliary data is already stored under the given name.
        """
        name = str_cast(
            name,
            'The name argument must be cast-able to type str!')

        if name in self._aux_data:
            raise KeyError(
                f'The auxiliary data "{name}" is already defined for dataset '
                f'"{self.name}"!')

        self._aux_data[name] = data

    def get_aux_data(
            self,
            name,
            default=None,
    ):
        """Retrieves the auxiliary data that is stored in this data set under
        the given name.

        Parameters
        ----------
        name : str
            The name under which the auxiliary data is stored.
        default : any | None
            If not ``None``, it specifies the returned default value when the
            auxiliary data does not exists.

        Returns
        -------
        data : unspecified
            The retrieved auxiliary data.

        Raises
        ------
        KeyError
            If no auxiliary data is stored with the given name and no default
            value was specified.
        """
        name = str_cast(
            name,
            'The name argument must be cast-able to type str!')

        if name not in self._aux_data:
            if default is not None:
                return default
            raise KeyError(
                f'The auxiliary data "{name}" is not defined for dataset '
                f'"{self.name}"!')

        return self._aux_data[name]

    def remove_aux_data(
            self,
            name,
    ):
        """Removes the auxiliary data that is stored in this data set under
        the given name.

        Parameters
        ----------
        name : str
            The name of the dataset that should get removed.
        """
        if name not in self._aux_data:
            raise KeyError(
                f'The auxiliary data "{name}" is not defined for dataset '
                f'"{self.name}", nothing to remove!')

        self._aux_data.pop(name)


class DatasetCollection(
        object):
    """The DatasetCollection class describes a collection of different datasets.

    New datasets can be added via the add-assign operator (+=), which calls
    the ``add_datasets`` method.
    """
    def __init__(
            self,
            name,
            description=''):
        """Creates a new DatasetCollection instance.

        Parameters
        ----------
        name : str
            The name of the collection.
        description : str
            The (longer) description of the dataset collection.
        """
        self.name = name
        self.description = description

        self._datasets = dict()

    @property
    def name(self):
        """The name (str) of the dataset collection.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name of the dataset collection must be of type str!')
        self._name = name

    @property
    def description(self):
        """The (longer) description of the dataset collection.
        """
        return self._description

    @description.setter
    def description(self, description):
        if not isinstance(description, str):
            raise TypeError(
                'The description of the dataset collection must be of type '
                'str!')
        self._description = description

    @property
    def dataset_names(self):
        """The list of names of the assigned datasets.
        """
        return sorted(self._datasets.keys())

    @property
    def version(self):
        """(read-only) The version number of the datasets collected by this
        dataset collection.
        """
        ds_name = list(self._datasets.keys())[0]
        return self._datasets[ds_name].version

    @property
    def verqualifiers(self):
        """(read-only) The dictionary holding the version qualifiers of the
        datasets collected by this dataset collection.
        """
        ds_name = list(self._datasets.keys())[0]
        return self._datasets[ds_name].verqualifiers

    def __getitem__(
            self,
            key,
    ):
        """Implementation of the access operator ``[key]``.

        Parameters
        ----------
        key : str | sequence of str
            The name or names of the dataset(s) that should get retrieved from
            this dataset collection.

        Returns
        -------
        datasets : instance of Dataset | list of instance of Dataset
            The dataset instance or the list of dataset instances corresponding
            to the given key.
        """
        if not issequence(key):
            return self.get_dataset(key)

        if not issequenceof(key, str):
            raise TypeError(
                'The key for the access operator must be an instance of str or '
                'a sequence of str instances!')

        datasets = [
            self.get_dataset(name)
            for name in key
        ]

        return datasets

    def __iadd__(self, ds):
        """Implementation of the ``self += dataset`` and
        ``self += (dataset1, dataset2, ...)`` operations to add one or several
        Dataset objects to this dataset collection.
        """
        self.add_datasets(ds)

        return self

    def __str__(self):
        """Implementation of the pretty string representation of the
        DatasetCollection instance. It shows the available datasets.
        """
        lines = f'DatasetCollection "{self.name}"\n'
        lines += "-"*display.PAGE_WIDTH + "\n"
        lines += "Description:\n" + self.description + "\n"
        lines += "Available datasets:\n"

        for name in self.dataset_names:
            lines += '\n'
            lines += display.add_leading_text_line_padding(
                2, str(self._datasets[name]))

        return lines

    def add_aux_data(
            self,
            name,
            data,
    ):
        """Adds the given data as auxiliary data to all datasets of this
        dataset collection.

        Parameters
        ----------
        name : str
            The name under which the auxiliary data will be stored.
        data : unspecified
            The data that should get stored. This can be of any data type.

        Raises
        ------
        ValueError
            If no datasets have been added to this dataset collection yet.
        KeyError
            If auxiliary data is already stored under the given name.
        """
        if len(self._datasets) == 0:
            raise ValueError(
                f'The dataset collection "{self.name}" has no datasets added '
                'yet!')

        for dataset in self._datasets.values():
            dataset.add_aux_data(
                name=name,
                data=data)

    def add_datasets(
            self,
            datasets,
    ):
        """Adds the given Dataset object(s) to this dataset collection.

        Parameters
        ----------
        datasets : instance of Dataset | sequence of instance of Dataset
            The instance of Dataset or the sequence of instance of Dataset that
            should be added to the dataset collection.

        Returns
        -------
        self : instance of DatasetCollection
            This instance of DatasetCollection in order to be able to chain
            several ``add_datasets`` calls.
        """
        if not issequence(datasets):
            datasets = [datasets]

        for dataset in datasets:
            if not isinstance(dataset, Dataset):
                raise TypeError(
                    'The dataset object must be a sub-class of Dataset!')

            if dataset.name in self._datasets:
                raise KeyError(
                    f'Dataset "{dataset.name}" already exists!')

            self._datasets[dataset.name] = dataset

        return self

    def remove_dataset(
            self,
            name,
    ):
        """Removes the given dataset from the collection.

        Parameters
        ----------
        name : str
            The name of the dataset that should get removed.
        """
        if name not in self._datasets:
            raise KeyError(
                f'Dataset "{name}" is not part of the dataset collection '
                f'"{self.name}", nothing to remove!')

        self._datasets.pop(name)

    def get_dataset(
            self,
            name,
    ):
        """Retrieves a Dataset object from this dataset collection.

        Parameters
        ----------
        name : str
            The name of the dataset.

        Returns
        -------
        dataset : Dataset instance
            The Dataset object holding all the information about the dataset.

        Raises
        ------
        KeyError
            If the data set of the given name is not present in this data set
            collection.
        """
        if name not in self._datasets:
            ds_names = '", "'.join(self.dataset_names)
            ds_names = '"'+ds_names+'"'
            raise KeyError(
                f'The dataset "{name}" is not part of the dataset collection '
                f'"{self.name}"! Possible dataset names are: {ds_names}!')

        return self._datasets[name]

    def get_datasets(
            self,
            names,
    ):
        """Retrieves a list of Dataset objects from this dataset collection.

        Parameters
        ----------
        names : str | sequence of str
            The name or sequence of names of the datasets to retrieve.

        Returns
        -------
        datasets : list of Dataset instances
            The list of Dataset instances for the given list of data set names.

        Raises
        ------
        KeyError
            If one of the requested data sets is not present in this data set
            collection.
        """
        if not issequence(names):
            names = [names]
        if not issequenceof(names, str):
            raise TypeError(
                'The names argument must be an instance of str or a sequence '
                'of str instances!')

        datasets = []
        for name in names:
            datasets.append(self.get_dataset(name))

        return datasets

    def set_exp_field_name_renaming_dict(
            self,
            d,
    ):
        """Sets the dictionary with the data field names of the experimental
        data that needs to be renamed just after loading the data. The
        dictionary will be set to all added data sets.

        Parameters
        ----------
        d : dict
            The dictionary with the old field names as keys and the new field
            names as values.
        """
        for (dsname, dataset) in self._datasets.items():
            dataset.exp_field_name_renaming_dict = d

    def set_mc_field_name_renaming_dict(
            self,
            d,
    ):
        """Sets the dictionary with the data field names of the monte-carlo
        data that needs to be renamed just after loading the data. The
        dictionary will be set to all added data sets.

        Parameters
        ----------
        d : dict
            The dictionary with the old field names as keys and the new field
            names as values.
        """
        for (dsname, dataset) in self._datasets.items():
            dataset.mc_field_name_renaming_dict = d

    def set_dataset_prop(
            self,
            name,
            value,
    ):
        """Sets the given property to the given name for all data sets of this
        data set collection.

        Parameters
        ----------
        name : str
            The name of the property.
        value : object
            The value to set for the given property.

        Raises
        ------
        KeyError
            If the given property does not exist in the data sets.
        """
        for (dsname, dataset) in self._datasets.items():
            if not hasattr(dataset, name):
                raise KeyError(
                    f'The dataset "{dsname}" does not have a property named '
                    f'"{name}"!')
            setattr(dataset, name, value)

    def define_binning(
            self,
            name,
            binedges,
    ):
        """Defines a binning definition and adds it to all the datasets of this
        dataset collection.

        Parameters
        ----------
        name : str
            The name of the binning definition.
        binedges : sequence
            The sequence of the bin edges, that should be used for the binning.
        """
        for dataset in self._datasets.values():
            dataset.define_binning(name, binedges)

    def add_data_preparation(
            self,
            func,
    ):
        """Adds the data preparation function to all the datasets of this
        dataset collection.

        Parameters
        ----------
        func : callable
            The object with call signature ``__call__(data)`` that will prepare
            the data after it was loaded. The argument 'data' is the DatasetData
            instance holding the experimental and monte-carlo data.
            This function must alter the properties of the DatasetData instance.
        """
        for dataset in self._datasets.values():
            dataset.add_data_preparation(func)

    def remove_data_preparation(
            self,
            key=-1,
    ):
        """Removes data preparation function from all the datasets of this
        dataset collection.

        Parameters
        ----------
        key : str, int, optional
            The name or the index of the data preparation function that should
            be removed. Default value is ``-1``, i.e. the last added function.

        Raises
        ------
        TypeError
            If the type of the key argument is invalid.
        IndexError
            If the given key is out of range.
        KeyError
            If the data preparation function cannot be found.
        """
        for dataset in self._datasets.values():
            dataset.remove_data_preparation(key=key)

    def update_version_qualifiers(
            self,
            verqualifiers,
    ):
        """Updates the version qualifiers of all datasets of this dataset
        collection.
        """
        for dataset in self._datasets.values():
            dataset.update_version_qualifiers(verqualifiers)

    def load_data(
            self,
            livetime=None,
            tl=None,
            ppbar=None,
            **kwargs,
    ):
        """Loads the data of all data sets of this data set collection.

        Parameters
        ----------
        livetime : float | dict of str => float | None
            If not None, uses this livetime (in days) as livetime for (all) the
            DatasetData instances, otherwise uses the live time from the Dataset
            instance. If a dictionary of data set names and floats is given, it
            defines the livetime for the individual data sets.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time the data load
            operation.
        ppbar : instance of ProgressBar | None
            The optional parent progress bar.
        **kwargs
            Additional keyword arguments are passed to the
            :meth:`~skyllh.core.dataset.Dataset.load_data` method of the
            individual datasets.

        Returns
        -------
        data_dict : dictionary str => instance of DatasetData
            The dictionary with the DatasetData instance holding the data of
            an individual data set as value and the data set's name as key.
        """
        if not isinstance(livetime, dict):
            livetime_dict = dict()
            for (dsname, dataset) in self._datasets.items():
                livetime_dict[dsname] = livetime
            livetime = livetime_dict

        if len(livetime) != len(self._datasets):
            raise ValueError(
                'The livetime argument must be None, a single float, or a '
                f'dictionary with {len(self._datasets)} str:float entries! '
                f'Currently the dictionary has {len(livetime)} entries.')

        pbar = ProgressBar(len(self._datasets), parent=ppbar).start()
        data_dict = dict()
        for (dsname, dataset) in self._datasets.items():
            data_dict[dsname] = dataset.load_data(
                livetime=livetime[dsname],
                tl=tl,
                **kwargs)
            pbar.increment()
        pbar.finish()

        return data_dict


class DatasetData(
        object):
    """This class provides the container for the actual experimental and
    monte-carlo data.
    """
    def __init__(
            self,
            data_exp,
            data_mc,
            livetime,
            **kwargs,
    ):
        """Creates a new DatasetData instance.

        Parameters
        ----------
        data_exp : instance of DataFieldRecordArray | None
            The instance of DataFieldRecordArray holding the experimental data.
            This can be None for a MC-only study.
        data_mc : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the monte-carlo data.
        livetime : float
            The integrated livetime in days of the data.
        """
        super().__init__(**kwargs)

        self.exp = data_exp
        self.mc = data_mc
        self.livetime = livetime

    @property
    def exp(self):
        """The DataFieldRecordArray instance holding the experimental data.
        This is None, if there is no experimental data available.
        """
        return self._exp

    @exp.setter
    def exp(self, data):
        if data is not None:
            if not isinstance(data, DataFieldRecordArray):
                raise TypeError(
                    'The exp property must be an instance of '
                    'DataFieldRecordArray!')
        self._exp = data

    @property
    def mc(self):
        """The DataFieldRecordArray instance holding the monte-carlo data.
        This is None, if there is no monte-carlo data available.
        """
        return self._mc

    @mc.setter
    def mc(self, data):
        if data is not None:
            if not isinstance(data, DataFieldRecordArray):
                raise TypeError(
                    'The mc property must be an instance of '
                    'DataFieldRecordArray!')
        self._mc = data

    @property
    def livetime(self):
        """The integrated livetime in days of the data.
        This is None, if there is no live-time provided.
        """
        return self._livetime

    @livetime.setter
    def livetime(self, lt):
        if lt is not None:
            lt = float_cast(
                lt,
                'The livetime property must be cast-able to type float!')
        self._livetime = lt

    @property
    def exp_field_names(self):
        """(read-only) The list of field names present in the experimental data.
        This is an empty list if there is no experimental data available.
        """
        if self._exp is None:
            return []
        return self._exp.field_name_list

    @property
    def mc_field_names(self):
        """(read-only) The list of field names present in the monte-carlo data.
        """
        return self._mc.field_name_list


def assert_data_format(
        dataset,
        data,
):
    """Checks the format of the experimental and monte-carlo data.

    Parameters
    ----------
    dataset : instance of Dataset
        The instance of Dataset describing the dataset and holding the local
        configuration.
    data : instance of DatasetData
        The instance of DatasetData holding the actual experimental and
        simulation data of the data set.

    Raises
    ------
    KeyError
        If a required data field is missing.
    """
    cfg = dataset.cfg

    def _get_missing_keys(keys, required_keys):
        missing_keys = []
        for reqkey in required_keys:
            if reqkey not in keys:
                missing_keys.append(reqkey)
        return missing_keys

    if data.exp is not None:
        missing_exp_keys = _get_missing_keys(
            data.exp.field_name_list,
            DataFields.get_joint_names(
                datafields=cfg['datafields'],
                stages=(
                    DFS.ANALYSIS_EXP
                )
            )
        )
        if len(missing_exp_keys) != 0:
            raise KeyError(
                'The following data fields are missing for the experimental '
                f'data of dataset "{dataset.name}": '
                ', '.join(missing_exp_keys))

    if data.mc is not None:
        missing_mc_keys = _get_missing_keys(
            data.mc.field_name_list,
            DataFields.get_joint_names(
                datafields=cfg['datafields'],
                stages=(
                    DFS.ANALYSIS_EXP |
                    DFS.ANALYSIS_MC
                )
            )
        )
        if len(missing_mc_keys) != 0:
            raise KeyError(
                'The following data fields are missing for the monte-carlo '
                f'data of dataset "{dataset.name}": '
                ', '.join(missing_mc_keys))

    if data.livetime is None:
        raise ValueError(
            f'No livetime was specified for dataset "{dataset.name}"!')


def remove_events(
        data_exp,
        mjds,
):
    """Utility function to remove events having the specified MJD time stamps.

    Parameters
    ----------
    data_exp : instance of DataFieldRecordArray
        The instance of DataFieldRecordArray holding the experimental data
        events.
    mjds : float | array of floats
        The MJD time stamps of the events, that should get removed from the
        experimental data array.

    Returns
    -------
    data_exp : instance of DataFieldRecordArray
        The instance of DataFieldRecordArray holding the experimental data
        events with the specified events removed.
    """
    mjds = np.atleast_1d(mjds)

    mask = np.zeros((len(data_exp)), dtype=np.bool_)
    for time in mjds:
        m = data_exp['time'] == time
        if np.count_nonzero(m) > 1:
            raise LookupError(
                f'The MJD time stamp {time} is not unique!')
        mask |= m
    data_exp = data_exp[~mask]

    return data_exp


def generate_base_path(
        default_base_path,
        base_path=None,
):
    """Generates the base path. If base_path is None, default_base_path is used.

    Parameters
    ----------
    default_base_path : str
        The default base path if base_path is None.
    base_path : str | None
        The user-specified base path.

    Returns
    -------
    base_path : str
        The generated base path.
    """
    if base_path is None:
        if default_base_path is None:
            raise ValueError(
                'The default_base_path argument must not be None, when the '
                'base_path argument is set to None!')
        base_path = default_base_path

    return base_path


def generate_sub_path(
        sub_path_fmt,
        version,
        verqualifiers,
):
    """Generates the sub path of the dataset based on the given sub path format.

    Parameters
    ----------
    sub_path_fmt : str
        The format string of the sub path.
    version : int
        The version of the dataset.
    verqualifiers : dict
        The dictionary holding the version qualifiers of the dataset.

    Returns
    -------
    sub_path : str
        The generated sub path.
    """
    fmt_dict = dict(
        [('version', version)] + list(verqualifiers.items())
    )
    sub_path = sub_path_fmt.format(**fmt_dict)

    return sub_path


def generate_data_file_root_dir(
        default_base_path,
        default_sub_path_fmt,
        version,
        verqualifiers,
        base_path=None,
        sub_path_fmt=None,
):
    """Generates the root directory of the data files based on the given base
    path and sub path format. If base_path is None, default_base_path is used.
    If sub_path_fmt is None, default_sub_path_fmt is used.

    The ``default_sub_path_fmt`` and ``sub_path_fmt`` arguments can contain the
    following wildcards:

        ``{version:d}``
            The version integer number of the dataset.
        ``{<verqualifiers_key>:d}``
            The integer number of the specific version qualifier
            ``'verqualifiers_key'``.

    Parameters
    ----------
    default_base_path : str
        The default base path if base_path is None.
    default_sub_path_fmt : str
        The default sub path format if sub_path_fmt is None.
    version : int
        The version of the data sample.
    verqualifiers : dict
        The dictionary holding the version qualifiers of the data sample.
    base_path : str | None
        The user-specified base path.
    sub_path_fmt : str | None
        The user-specified sub path format.

    Returns
    -------
    root_dir : str
        The generated root directory of the data files. This will have no
        trailing directory separator.
    """
    base_path = generate_base_path(
        default_base_path=default_base_path,
        base_path=base_path)

    if sub_path_fmt is None:
        sub_path_fmt = default_sub_path_fmt

    sub_path = generate_sub_path(
        sub_path_fmt=sub_path_fmt,
        version=version,
        verqualifiers=verqualifiers)

    root_dir = os.path.join(base_path, sub_path)

    len_sep = len(os.path.sep)
    if root_dir[-len_sep:] == os.path.sep:
        root_dir = root_dir[:-len_sep]

    return root_dir


def get_data_subset(
        data,
        livetime,
        t_start,
        t_stop,
):
    """Gets instance of DatasetData and instance of Livetime with data subsets
    between the given time range from ``t_start`` to ``t_stop``.

    Parameters
    ----------
    data : DatasetData
        The DatasetData object.
    livetime : Livetime
        The Livetime object.
    t_start : float
        The MJD start time of the time range to consider.
    t_end : float
        The MJD end time of the time range to consider.

    Returns
    -------
    data_subset : instance of DatasetData
        The instance of DatasetData with subset of the data between the given
        time range from ``t_start`` to ``t_stop``.
    livetime_subset : instance of Livetime
        The instance of Livetime for a subset of the data between the given
        time range from ``t_start`` to ``t_stop``.
    """
    if not isinstance(data, DatasetData):
        raise TypeError(
            'The "data" argument must be of type DatasetData!')
    if not isinstance(livetime, Livetime):
        raise TypeError(
            'The "livetime" argument must be of type Livetime!')

    exp_slice = np.logical_and(
        data.exp['time'] >= t_start,
        data.exp['time'] < t_stop)
    mc_slice = np.logical_and(
        data.mc['time'] >= t_start,
        data.mc['time'] < t_stop)

    data_exp = data.exp[exp_slice]
    data_mc = data.mc[mc_slice]

    uptime_mjd_intervals_arr = livetime.get_uptime_intervals_between(
        t_start, t_stop)
    livetime_subset = Livetime(uptime_mjd_intervals_arr)

    data_subset = DatasetData(
        data_exp=data_exp,
        data_mc=data_mc,
        livetime=livetime_subset.livetime)

    return (data_subset, livetime_subset)



# -*- coding: utf-8 -*-

"""The ``backgroundpdf`` module contains background PDF classes for the
likelihood function.
"""

import numpy as np

from skyllh.core.pdf import (
    IsBackgroundPDF,
    MultiDimGridPDF,
    TimePDF,
)
from skyllh.core.py import (
    classname,
)


class BackgroundMultiDimGridPDF(
        MultiDimGridPDF,
        IsBackgroundPDF):
    """This class provides a multi-dimensional background PDF defined on a grid.
    The PDF is created from pre-calculated PDF data on a grid. The grid data is
    interpolated using a :class:`scipy.interpolate.RegularGridInterpolator`
    instance.
    """

    def __init__(
            self,
            *args,
            **kwargs):
        """Creates a new :class:`~skyllh.core.pdf.MultiDimGridPDF` instance that
        is also derived from :class:`~skyllh.core.pdf.IsBackgroundPDF`.

        For the documentation of arguments see the documentation of the
        :meth:`~skyllh.core.pdf.MultiDimGridPDF.__init__` method.
        """
        super().__init__(*args, **kwargs)


class BackgroundTimePDF(
        TimePDF,
        IsBackgroundPDF):
    """This class provides a background time PDF class.
    """

    def __init__(
            self,
            livetime,
            time_flux_profile,
            **kwargs):
        """Creates a new signal time PDF instance for a given time flux profile
        and detector live time.

        Parameters
        ----------
        livetime : instance of Livetime
            An instance of Livetime, which provides the detector live-time
            information.
        time_flux_profile : instance of TimeFluxProfile
            The signal's time flux profile.
        """
        super().__init__(
            pmm=None,
            livetime=livetime,
            time_flux_profile=time_flux_profile,
            **kwargs)

        self._pd = None

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Initializes the background time PDF with new trial data. Because this
        PDF does not depend on any parameters, the probability density values
        can be pre-computed here.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF value. The following data fields must
            exist:

            ``'time'`` : float
                The MJD time of the event.

        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.
        """
        times = tdm.get_data('time')

        self._pd = np.zeros((len(times),), dtype=np.float64)

        # Get a mask of the event times which fall inside a detector on-time
        # interval.
        on = self._livetime.is_on(times)

        self._pd[on] = self._time_flux_profile(t=times[on]) / self._S

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """
        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF value. The following data fields must
            exist:

            ``'time'`` : float
                The MJD time of the event.

        params_recarray : None
            Unused interface argument.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_events,)-shaped numpy ndarray holding the background
            probability density value for each event.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter.
            The background PDF does not depend on any global fit parameter,
            hence, this is an empty dictionary.
        """
        if self._pd is None:
            raise RuntimeError(
                f'The {classname(self)} was not initialized with trial data!')

        grads = dict()

        return (self._pd, grads)



# -*- coding: utf-8 -*-

import abc

from skyllh.core.background_generation import (
    BackgroundGenerationMethod,
)
from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.py import (
    classname,
    issequenceof,
)
from skyllh.core.random import (
    RandomStateService,
)
from skyllh.core.timing import (
    TaskTimer,
)


class BackgroundGenerator(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """This is the abstract base class for all background generator classes in
    SkyLLH. It defines the interface for a background generator.
    """
    def __init__(
            self,
            **kwargs,
    ):
        """Constructs a new instance of BackgroundGenerator.

        Parameters
        ----------
        bkg_gen_method : instance of BackgroundGenerationMethod
            The optional background event generation method, which should be
            used to generate events.
        """
        super().__init__(**kwargs)

    def change_shg_mgr(self, shg_mgr):
        """This method should be reimplemented when the background generator
        depends on the sources.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        pass

    @abc.abstractmethod
    def generate_background_events(
            self,
            rss,
            tl=None,
            **kwargs,
    ):
        """This method is supposed to generate a mean number of background
        events for the datasets of this background generator.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which will be passed to the
            ``generate_events`` method of the background generation method
            instance. Usual keyword arguments are `mean`, the mean number of
            background events to generate, and `poisson`, the flag if the number
            of background events should get drawn from a Poisson distribution
            with the given mean number of background events as mean.

        Returns
        -------
        n_bkg_list : list of int
            The number of generated background events.
        bkg_events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray holding the generated
            background events. The number of events can be less than stated in
            `n_bkg_list` if an event selection method is used.
        """
        pass


class DatasetBackgroundGenerator(
        BackgroundGenerator,
):
    """This class provides a background generator for a particular dataset.
    It holds a background generation method which is used to generate the
    background events.
    """

    def __init__(
            self,
            dataset,
            data,
            bkg_gen_method,
            **kwargs,
    ):
        super().__init__(**kwargs)

        self.dataset = dataset
        self.data = data
        self.bkg_gen_method = bkg_gen_method

    @property
    def dataset(self):
        """The instance of Dataset for which background events should get
        generated.
        """
        return self._dataset

    @dataset.setter
    def dataset(self, ds):
        if not isinstance(ds, Dataset):
            raise TypeError(
                'The dataset property must be an instance of Dataset! '
                f'Its current type is {classname(ds)}.')
        self._dataset = ds

    @property
    def data(self):
        """The instance of DatasetData holding the experimental and simulation
        data of the dataset.
        """
        return self._data

    @data.setter
    def data(self, d):
        if not isinstance(d, DatasetData):
            raise TypeError(
                'The data property must be an instance of DatasetData! '
                f'Its current type is {classname(d)}.')
        self._data = d

    @property
    def bkg_gen_method(self):
        """The instance of BackgroundGenerationMethod which should be used to
        generate background events. This can be ``None``.
        """
        return self._bkg_gen_method

    @bkg_gen_method.setter
    def bkg_gen_method(self, method):
        if method is not None:
            if not isinstance(method, BackgroundGenerationMethod):
                raise TypeError(
                    'The bkg_gen_method property must be an instance of '
                    'BackgroundGenerationMethod! '
                    f'Its current type is {classname(method)}.')
        self._bkg_gen_method = method

    def change_shg_mgr(self, shg_mgr):
        """Changes the SourceHypoGroupManager instance of the background
        generation method.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        self._bkg_gen_method.change_shg_mgr(
            shg_mgr=shg_mgr)

    def generate_background_events(
            self,
            rss,
            tl=None,
            **kwargs,
    ):
        """Generates a mean number of background events for the dataset of this
        background generator.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which will be passed to the
            ``generate_events`` method of the background generation method
            instance. Usual keyword arguments are `mean`, the mean number of
            background events to generate, and `poisson`, the flag if the number
            of background events should get drawn from a Poisson distribution
            with the given mean number of background events as mean.

        Returns
        -------
        n_bkg_list : list of int
            The list of length 1 holding the number of generated background
            events for the dataset.
        bkg_events_list : list of instance of DataFieldRecordArray
            The list of length 1 holding the instance of DataFieldRecordArray
            holding the generated background events. The number of events can be
            less than stated in ``n_bkg_list`` if an event selection method is
            used.
        """
        (n_bkg, bkg_events) = self._bkg_gen_method.generate_events(
            rss=rss,
            dataset=self._dataset,
            data=self._data,
            tl=tl,
            **kwargs)

        return ([n_bkg], [bkg_events])


class MultiDatasetBackgroundGenerator(
        BackgroundGenerator,
):
    """This is a background generator class handling multiple datasets by using
    the individual background generator instances for each dataset. This is the
    most general way to support multiple datasets of different formats and
    background generation.
    """

    def __init__(
            self,
            dataset_list,
            data_list,
            bkg_generator_list,
            **kwargs,
    ):
        """Constructs a new instance of MultiDatasetBackgroundGenerator.

        Parameters
        ----------
        dataset_list : list of instance of Dataset
            The list of Dataset instances for which background events should get
            generated for.
        data_list : list of instance of DatasetData
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        bkg_generator_list : list of instance of BackgroundGenerator
            The list of BackgroundGenerator instances, one for each dataset.
            The order must match the order of ``dataset_list``.
        """
        super().__init__(
            **kwargs)

        self.dataset_list = dataset_list
        self.data_list = data_list
        self.bkg_generator_list = bkg_generator_list

    @property
    def dataset_list(self):
        """The list of Dataset instances for which background events should get
        generated for.
        """
        return self._dataset_list

    @dataset_list.setter
    def dataset_list(self, datasets):
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The dataset_list property must be a sequence of Dataset '
                'instances! '
                f'Its current type is {classname(datasets)}.')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the actual data of each
        dataset. The order must match the order of the ``dataset_list``
        property.
        """
        return self._data_list

    @data_list.setter
    def data_list(self, datas):
        if not issequenceof(datas, DatasetData):
            raise TypeError(
                'The data_list property must be a sequence of DatasetData '
                'instances! '
                f'Its current type is {classname(datas)}.')
        self._data_list = datas

    @property
    def bkg_generator_list(self):
        """The list of instance of BackgroundGenerator, one for each dataset.
        The order must match the order of the ``dataset_list`` property.
        """
        return self._bkg_generator_list

    @bkg_generator_list.setter
    def bkg_generator_list(self, generators):
        if not issequenceof(generators, BackgroundGenerator):
            raise TypeError(
                'The bkg_generator_list property must be a sequence of '
                'BackgroundGenerator instances! '
                f'Its current type is {classname(generators)}.')
        self._bkg_generator_list = generators

    def change_shg_mgr(
            self,
            shg_mgr,
    ):
        """Calls the ``change_shg_mgr`` method of each individual dataset
        background generator.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        for bkg_generator in self._bkg_generator_list:
            bkg_generator.change_shg_mgr(
                shg_mgr=shg_mgr)

    def generate_background_events(
            self,
            rss,
            mean_n_bkg_list=None,
            tl=None,
            **kwargs,
    ):
        """Generates a mean number of background events for each individual
        dataset of this multi-dataset background generator.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the individual
            background generator instance needs to obtain this number itself.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which will be passed to the
            ``generate_background_events`` method of the individual background
            generator instance.

        Returns
        -------
        n_bkg_events_list : list of int
            The list holding the number of generated background events for each
            dataset.
        bkg_events_list : list of instance of DataFieldRecordArray
            The list holding the instance of DataFieldRecordArray holding the
            generated background events of each dataset. The number of events
            can be less than stated in ``n_bkg_list`` if an event selection
            method is used.
        """
        if not isinstance(rss, RandomStateService):
            raise TypeError(
                'The rss argument must be an instance of RandomStateService! '
                f'Its current type is {classname(rss)}.')

        if mean_n_bkg_list is None:
            mean_n_bkg_list = [None] * len(self._bkg_generator_list)
        if not issequenceof(mean_n_bkg_list, (type(None), float)):
            raise TypeError(
                'The mean_n_bkg_list argument must be a sequence of None '
                'and/or floats! '
                f'Its current type is {classname(mean_n_bkg_list)}.')

        if kwargs is None:
            kwargs = dict()

        n_bkg_events_list = []
        bkg_events_list = []
        for (ds, bkg_generator, mean_n_bkg) in zip(
                self._dataset_list, self._bkg_generator_list, mean_n_bkg_list):
            kwargs.update(mean=mean_n_bkg)
            with TaskTimer(
                    tl,
                    f'Generating background events for dataset "{ds.name}".'):
                (n_bkg_events_list_, bkg_events_list_) =\
                    bkg_generator.generate_background_events(
                        rss=rss,
                        tl=tl,
                        **kwargs)
            n_bkg_events_list += n_bkg_events_list_
            bkg_events_list += bkg_events_list_

        return (n_bkg_events_list, bkg_events_list)
